// -*- Mode:C++ -*-
//Header:
//File: genTraj.cc
//Title: Saloo 
//Author: NODA, Itsuki
//Date: 2001/03/22
//

//ModifyHistory:
// 2001/03/22: Start to create this file
//EndModifyHistory:


#include "SimplePlayer.h"
#include "LogTrajPlayerT.h"

using namespace Saloo ;
using namespace Itk ;

//======================================================================
// parameters

class Param {
   public:
      SString logfile ;
      Int loopMax ;
} ;

Param MyParam = {"foo.trj",
		 10000} ;

//======================================================================
// GenTrajPlayer2 class

class GenTrajPlayer2 : public LogTrajPlayerT<SimplePlayer> {
   public:
      virtual Bool turn(const Flt & moment) ;

//      virtual Bool recv();
      ostream * _dump ;
      ITK_DEF_ACCESS(ostream & dump(), { return *_dump ; }) ;
      GenTrajPlayer2() {
	 _dump = new ofstream("foo.recv.dump") ;
      } ;
	 
} ;

Bool GenTrajPlayer2::turn(const Flt & moment) {
   if(Abs(moment) < 1.0) return True ;
   else return LogTrajPlayer::turn(moment) ;
} ;

/*
Bool GenTrajPlayer2::recv(){
   Bool r = LogTrajPlayer::recv() ;
   if(r) 
      dump() << recvbuf << endl ;
   return r ;
} ;
*/

//======================================================================
// main loop

Bool mainloop(Param & param) {
   GenTrajPlayer2 player ;

   player.init() ;
   player.openLog(param.logfile) ;

   player.forkLogCoachSensor() ;
   player.forkLogPlayerSensor() ;

   player.move(-5,0) ;
   player.turn(180);
//   player.turn(-90) ;
   player.control.sendChangeMode(PLAYMODE(PlayOn)) ;

   enum ActMode {
      M_Random,
      M_GoCenter,
   } mode ;
   mode = M_Random ;
   
   while(True) {
      if(param.loopMax >= 0 && 
	 player.stime().count > static_cast<UInt>(param.loopMax)) {
	 player.closeLog(True) ;  // die in this function
      }
      if(player.c_kickable()) {
	 if(mode == M_GoCenter) {
	    if(player.c_mypos().norm() < 9.0) {
	       mode = M_Random ;
	    } else {
	       player.kickToAbs(fltRand(5,100),Pos(0,0)) ;
	    }
	 } 
	 else { // mode == M_Random
	    if(player.c_mypos().norm() > 30.0) {
	       mode = M_GoCenter ;
	    } else {
	       Int rnd = intRand(100) ;
	       Int thr = 90 ;
	       if(rnd < thr) {
		  if(rnd < (thr/2)) {
		     player.kick(fltRand(1,10),fltRand(-180,-90)) ;
		  } else {
		     player.kick(fltRand(1,10),fltRand(90,180)) ;
		  }
	       } else {
		  player.kick(fltRand(1,100),fltRand(-180,180)) ;
	       }
	    }
	 }
      }
      else {
	 player.turnToAbs(player.c_ballpos()) ;
	 player.dash(fltRand(10,100)) ;
      }
   }

   return True ;
}

//======================================================================
// getOpt

Bool getOpt(int & argc, char ** & argv, Param & param) {
   //----------------------------------------
   OptArg optlist[] = {

      {"logfile",	OptArg::SStringVal,	&(param.logfile),	0,
       "-logfile LOGFILENAME"},

      {"loopMax",	OptArg::IntVal,		&(param.loopMax),	0,
       "-loopMax LOOPMAX"},
   
      {ITK_NULLPTR}} ;
   //----------------------------------------
   
   myGetOpt(optlist, argc, argv) ;

   return True ;
}

//======================================================================
// main

int main(int argc, char** argv) {
   getOpt(argc,argv,MyParam) ;

   mainloop(MyParam) ;

   return 0 ;
} ;




	 


