// -*- Mode:C++ -*-
//Header:
//File: Bun.h
//Title: Bun Player for chagamma
//Author: NODA, Itsuki
//Date: 2002/04/10
//

//ModifyHistory:
// 2001/03/02: Start to create this file
// 2002/04/10: Branch from chagama.cc
//EndModifyHistory:

#ifndef _Bun_Bun_h_
#define _Bun_Bun_h_
////////////////////////////////////////////////////////////////////////

#include "itk/itk.h"

#include <stl.h>
#include "SimplePlayer.h"
#include "OnlineCoach.h"
#include "OfflineCoach.h"

#include "Param.h"
#include "PlayPhase.h"

#include "Newton2dKFilter.h"

namespace Soccer {

//======================================================================
// class Bun

   /*--------------------*/
   /**
    *
    */

   class Bun : public SimplePlayer {

	 //============================================================
	 // Bun:: class TargetInfo
	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 class TargetInfo : public WithDescriber {

	       /*--------------------*/
	       /**
		*
		*/
	    public:
	       enum Type {
		  T_None,
		  T_Goal,
		  T_Clear,
		  T_Pass,
		  T_Dribble,
		  T_Max } ;

	    public:
	       Type type ;
	       Pos pos ;
	       Flt score ;
	       Flt power ;
	       char * name ;
	       Int aux ;

	    public:
	       virtual void describe(ostream & ostr, 
				     const Bool detailp = True) const {
		  ostr << "#TargetInfo[" << pos ;
		  switch(type) {
		     case T_None :		ostr << "(N)" ; break ;
		     case T_Goal :		ostr << "(G)" ; break ;
		     case T_Clear : 	ostr << "(C)" ; break ;
		     case T_Pass : 	ostr << "(P)" ; break ;
		     case T_Dribble : 	ostr << "(D)" ; break ;
		     default :       	ostr << "(?)" ;
		  }
		  ostr << "/power=" << power ;
		  ostr << ":score=" << score 
		       << "|" << name << "(" << aux << ")]" ;

		  if(detailp) ostr << endl ;
	       } ;
	 } ;

	 //============================================================
	 // Bun:: class CutPosInfo
	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 class CutPosInfo : public WithDescriber {
	    public:
	       Vec pos ;
	       Int maxdash ;

	       /* <<< [Bun-0.3] <<< */

	       Int endtime ;

	       /* >>> [Bun-0.3] >>> */

	       /*--------------------*/
	       /**
		*
		*/
	    public:
	       CutPosInfo() {} ;

	       /*--------------------*/
	       /**
		*
		*/
	    public:
	       CutPosInfo(const Vec & p, const Int & m, const Int & et=-1) {
		  pos = p ; maxdash = m ; 

		  /* <<< [Bun-0.3] <<< */

		  endtime = et ;

		  /* >>> [Bun-0.3] >>> */
	       } ;

	    public:
	       virtual void describe(ostream & ostr, 
				     const Bool detailp = True) const {
		  ostr << "#CutPosInfo[" << pos 
		       << ",maxdash=" << maxdash 
		       << ",endtime=" << endtime
		       << "]" ;
	       }
	 } ;

	 //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	 // Bun:: member
      public:
	 Param param ;
	 Param::RoleInfo * role ;

	 PlayPhaseStack ppstack ;

	 //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	 // Bun:: hear utility
      public:
	 SimpleSexp::Heap hearItemHeap ;
	 Mutex mtxHear ;

	 //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	 // Bun:: say process and utility
      public:
	 SimpleSexp::Heap sayItemHeap ;
	 SimpleSexp * saylist ;
	 Mutex mtxSay ;

	 Thread * thrdSay ;

	 /* <<< [Bun-0.4] <<< */

	 //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	 // Bun:: delayed keepLookBall info
      public:
	 Mutex mtxDelayedTurnHead ;
	 Bool needDelayedTurnHead ;

	 /* >>> [Bun-0.4] >>> */

	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 void clearSay(Bool lockp = False) {
	    if(lockp) mtxSay.lock() ;
	    {
	       sayItemHeap.clear() ;
	       saylist = SimpleSexp::Nil ;
	    }
	    if(lockp) mtxSay.unlock() ;
	 } ;

	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 SimpleSexp * addSayItem(SimpleSexp * s) {
	    return sayItemHeap.addFirst(s,saylist) ;
	 } ;

	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 Thread * forkSay() ;

	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 static void * runSay(void *p) ;

	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 void * sayMainLoop() ;

	 //------------------------------------------------------------
	 // Bun:: constructor
	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 Bun() { 
	 } ;

	 //------------------------------------------------------------
	 // Bun:: run 
	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 void run() ;

	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 void mainloop() ;

	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 void playDango() ;

	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 void playFixedPosition() ;
      
	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 void playGoalie() ;

	 //------------------------------------------------------------
	 // Bun:: test run
	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 void playTest1() ;
	 void playTest2() ;
	 void playTest3() ;
	 void playTest4() ;
	 void playTest5() ;
	 void playTest6() ;
	 void playTest7() ;

	 void playTest8() ;
	 void playTest8_1() ;
	 void playTest8_2() ;

	 /* <<< [Bun-0.3] <<< */
	 void playTest9() ;
	 void playTest9_1() ;
	 void playTest9_2() ;
	 /* >>> [Bun-0.3] >>> */


	 //------------------------------------------------------------
	 // Bun:: utility
	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 Bool checkKickOff() ;

	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 Bool checkPos() ;

	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 Bool checkBall() ;

	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 Flt badnessOfTarget(const Pos & target,TargetInfo::Type type) ;

	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 Flt goodnessOfTarget(const Pos & target,TargetInfo::Type type) ;

	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 Flt scoreOfTarget(const Pos & target, TargetInfo::Type type) ;

	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 Bool compareTargetScore(TargetInfo & target, const Pos & newpos,
				 char * name = ITK_NULLPTR, 
				 TargetInfo::Type type = TargetInfo::T_None,
				 Int aux = 0) ;

	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 TargetInfo findBestTargetPos() ; // used in playFixedPosition

	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 Bool shouldChase() ;      // used in playFixedPosition

	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 Bool isClosestToBall(Int predict = 0, Bool ignoreGoalie = True) ;

	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 Pos homepos() ;

	 //------------------------------------------------------------
	 // Bun:: goalie's utility
	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 Bool maybeSuccessCatch() {
	    if(isLeft(side)) {
	       return (playmode()->id == SC_WID_GoalieCatchL ||
		       playmode()->id == SC_WID_FreeKickL) ;
	    } else {
	       return (playmode()->id == SC_WID_GoalieCatchR ||
		       playmode()->id == SC_WID_FreeKickR) ;
	    }
	 } ;
      
	 //------------------------------------------------------------
	 // Bun:: find ball cut pos
	 /**
	  *
	  * dashN  dash 뤳ȤǤ pos ޤãǤ뤫ɤĴ٤롣
	  */
      public:
	 Bool canReachTo(const Pos & pos, Int c) ;

	 /**
	  *
	  * ball ã dash ǹԤ뤫ɤˤäơ 
	  * chase 
	  * ꤹ롣
	  */
      public:
	 //<<<[Bun-0.1]
	 //CutPosInfo cutBallPos() { return cutBallPos1() ; } ;
	 CutPosInfo cutBallPos() ;
	 //>>>[Bun-0.1]

	 //<<<[Bun-0.1]
	 //CutPosInfo cutBallPos0() ;
	 //CutPosInfo cutBallPos1() ;
	 //>>>[Bun-0.1]

	 //------------------------------------------------------------
	 // Bun:: exactRun : do exact run to the dest position
	 /*--------------------*/
	 /**
	  *
	  * useBackDashp  True ʤ顢٤ 90 ٰʾξ back dash 
	  * 롣
	  */
      public:
	 void exactRun(Pos destpos, Flt maxpower,
		       Bool needWaitFirstp = False, 
		       Int maxdash = 0,
		       Bool useBackDashp = False) ;

	 /* <<< [Bun-0.3] <<< */

	 /*--------------------*/
	 /**
	  *
	  * ֤ޤǤ˥ܡͽ¬ޤ롣
	  */
      public:
	 void chacingRun(Flt maxpower, Int endtime,
			  Bool needWaitFirstp = False) ; 

	 /* >>> [Bun-0.3] >>> */

	 /* <<< [Bun-0.4] <<< */

	 /*--------------------*/
	 /**
	  *
	  * ܡ뤫򤽤餵ʤ褦ˤʤ
	  * ư롣
	  */
      public:
	 void homingRun(Pos destpos, Flt maxpower) ;

	 /* >>> [Bun-0.4] >>> */

	 //------------------------------------------------------------
	 // Bun:: lookAroundWell
	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 void lookAroundWell() ;

	 /* <<< [Bun-0.4] <<< */

	 /*--------------------*/
	 /**
	  *
	  * Turn θ塢ܡ³褦Ĵ
	  * immediatep  True λˤľ turnHead 򤹤롣
	  * False λˤϡ sensebody 褿Ȥ˹Ԥ
	  */
      public:
	 //void keepLookingBall(Bool immediatep = False) ;
	 void keepLookingBall(Bool immediatep = True) ;

	 /*--------------------*/
	 /**
	  *
	  * ܡΤ˴롣
	  * predict ϡʬܡΰưͽ¬
	  */
      public:
	 void 
	 faceBall(Int predict = 0) ;

	 /* >>> [Bun-0.4] >>> */


	 //------------------------------------------------------------
	 // Bun:: PlayPhase manipuration

	 //----------------------------------------
	 // Bun:: push base
	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 PlayPhase & pushTopPlayPhase(Bool lockp = True) ;

	 //----------------------------------------
	 // Bun:: push content
	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 PlayPhase & pushPlayPhase(const PlayPhase & p, Bool lockp = True) ;
      
	 //----------------------------------------
	 // Bun:: pop
	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 PlayPhase & popPlayPhase(Bool lockp = True) ;

	 //----------------------------------------
	 // Bun:: push with Flt args
	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 PlayPhase & pushPlayPhase(PlayPhase::Type type, 
				   Flt arg1,
				   Flt arg2 = 0.0,
				   Flt arg3 = 0.0,
				   Bool lockp = True) ;

	 //----------------------------------------
	 // Bun:: push with Int args
	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 PlayPhase & pushPlayPhase(PlayPhase::Type type, 
				   Int arg1 = 0,
				   Int arg2 = 0,
				   Int arg3 = 0,
				   Bool lockp = True) ;

	 //----------------------------------------
	 // Bun:: get current PlayPhase
	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 PlayPhase & currentPlayPhase(Bool lockp = True) ;

	 //----------------------------------------
	 // Bun:: change current PlayPhase
	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 PlayPhase & changePlayPhase(const PlayPhase & newpp, Bool lockp = True) ;

	 //----------------------------------------
	 // Bun:: change current PlayPhase with Flt args
	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 PlayPhase & changePlayPhase(PlayPhase::Type type,
				     Flt arg1,
				     Flt arg2 = 0.0,
				     Flt arg3 = 0.0,
				     Bool lockp = True) ;

	 //----------------------------------------
	 // Bun:: change current PlayPhase with Int args
	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 PlayPhase & changePlayPhase(PlayPhase::Type type,
				     Int arg1 = 0,
				     Int arg2 = 0,
				     Int arg3 = 0,
				     Bool lockp = True) ;

	 //----------------------------------------
	 // Bun:: confirm or push current PlayPhase
	 // 	return True if new PlayPhase is pushed
	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 Bool confirmPushPlayPhase(PlayPhase::Type type,
				   Int arg1 = 0,
				   Int arg2 = 0,
				   Int arg3 = 0,
				   Bool lockp = True) ;

	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 Bool confirmPushPlayPhase(PlayPhase::Type type,
				   Flt arg1,
				   Flt arg2 = 0.0,
				   Flt arg3 = 0.0,
				   Bool lockp = True) ;
	    
	 //----------------------------------------
	 // Bun:: confirm and pop current PlayPhase
	 // 	return True if PlayPhase is poped
	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 Bool confirmPopPlayPhase(PlayPhase::Type type, Bool lockp = True) ;

	 //----------------------------------------
	 // Bun:: virtual analyzeSensorSee()
	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 virtual Bool analyzeSensorSee() ;

	 //----------------------------------------
	 // Bun:: virtual analyzeSensorHear() ;
	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 virtual Bool analyzeSensorHear() ;

	 /* <<< [Bun-0.4] <<< */

	 //----------------------------------------
	 // Bun:: virtual analyzeSensorHear() ;
	 /*--------------------*/
	 /**
	  *
	  * delayed turnHead 
	  */
      public:
	 virtual Bool analyzeSensorSenseBody() ;

	 /* >>> [Bun-0.4] >>> */

	 /* <<< [Bun-0.5] <<< */

	 //------------------------------------------------------------
	 // Bun:: for goalie

	 /*--------------------*/
	 /**
	  *
	  * 
	  */
      public:
	 void mainloopGoalie() ;
	 

	 /* >>> [Bun-0.5] >>> */

	 //----------------------------------------
	 // Bun:: for debug
	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 virtual Bool send() {
	    //ITK_MSG_T("sb:" << endl << "\t\t" << *this << btime().rtime) ;
	    //ITK_MSG_T("send:" << endl << "\t\t" << *this << TimeVal::now()) ;
	    
	    Bool r = SimplePlayer::send() ;

	    /* <<< [Bun-0.2] <<< */
	    if(False) { 
	    //if(unum == 1) { /* for debug */
	       if(!(sendbuf.hasPrefix("(say ") ||
		    sendbuf.hasPrefix("(turn_neck ") ||
		    sendbuf.hasPrefix("(change_view "))) {
		  ITK_MSG_T("send:" << *this << TimeVal::now() << endl
			    << "\t" << btime() << "\t" << sendbuf) ;
	       }
	    }
	    /* >>> [Bun-0.2] >>> */

	    return r ;
	 } ;

	 //----------------------------------------
	 // Bun:: Kalman Filter (for debug)
	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 Newton2dKFilter kfilter ;

   } ;
   
} ;
////////////////////////////////////////////////////////////////////////
#endif

