// -*- Mode:C++ -*-
//Header:
//File: FieldState.h
//Title: Field Status 
//Author: NODA, Itsuki
//Date: 2001/03/02
//

//ModifyHistory:
// 2001/03/02: Start to create this file
//EndModifyHistory:

#ifndef _SPlayer_FieldState_h_
#define _SPlayer_FieldState_h_
////////////////////////////////////////////////////////////////////////

#include "itk/itk.h"
#include "sctypes.h"
#include "CycleStep.h"
#include "SensorInfo.h"
#include "BasePlayer.h"

#ifdef OLD_ESTIMATOR
#include "estimatePos.h"
#else
#include "PosEstimator.h"
#endif

namespace Soccer {

   using namespace Itk ;

   //======================================================================
   // class FieldState

   /*--------------------*/
   /**
    *
    */

   class FieldState : public WithDescriber {

	 //--------------------------------------------------
	 // FieldState:: constant
	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 static const UInt playerN = SC_N_PLAYER_PER_TEAM ;

	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 static const UInt playerM = 2 * SC_N_PLAYER_PER_TEAM ;

	 //==================================================
	 // FieldState:: class Object
	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 class Object : public WithDescriber {

	       //------------------------------
	       // FieldState::Object:: Sensor Value
	       /*--------------------*/
	       /**
		*
		*/
	    public:
	       typedef SeeInfo::Object SensorValue ;

	       //------------------------------
	       // FieldState::Object:: Quality
	       /*--------------------*/
	       /**
		*
		*/
	    public:
	       enum Quality {
		  Q_None = SeeInfo::SQ_none,
		  Q_Dist = SeeInfo::SQ_dist,
		  Q_Dir = SeeInfo::SQ_dir,
		  Q_DistChng = SeeInfo::SQ_dist_chng,
		  Q_DirChng = SeeInfo::SQ_dir_chng,
		  Q_BodyDir = SeeInfo::SQ_body_dir,
		  Q_HeadDir = SeeInfo::SQ_head_dir
	       } ;

	       //------------------------------
	       // FieldState::Object::object type
	       //	is equal to SeeInfo::Object::Type
	       /*--------------------*/
	       /**
		*
		*/
	    public:
	       enum Type {
		  Self   = SensorValue::Self,
		  Ball   = SensorValue::Ball,
		  Player = SensorValue::Player,
		  Goal   = SensorValue::Goal,
		  Flag   = SensorValue::Flag,
		  Line   = SensorValue::Line,
		  Other  = SensorValue::Other
	       } ;

	       /*--------------------*/
	       /**
		*
		*/
	    public:
	       static const char * typestr(Type type) {
		  switch(type) {
		     case Self:		return "self" ;
		     case Ball: 	return "ball" ;
		     case Player: 	return "player" ;
		     case Goal:		return "goal" ;
		     case Flag:         return "flag" ;
		     case Line:		return "line" ;
		     case Other:	return "other" ;
		     default:		return "unknown" ;
		  }
	       } ;

	       //------------------------------
	       // FieldState::Object:: unknown unum
	       /*--------------------*/
	       /**
		*
		*/
	    public:
	       static const UInt UnknownUNum 
	       = SeeInfo::Object::PlayerID::UnknownUNum ;

	       //==============================
	       // FieldState::Object:: body (Object)

	       //------------------------------
	       // FieldState::Object:: member
	    public:
	       Type type ;
	       Int aux ;

	       CycleStep _time ;

	       PosDir _posdir ;
	       Flt posConf ;

	       Flt posVariant ; // ָʬ

	       Vec _vel ;
	       Flt velConf ;

	       /*--------------------*/
	       /**
		*
		* ǡμ¤ɤ򼨤
		* ˡball 줿Ȥʤɤˡ
		* ξ󤬳μ¤ǤʤΤʤȤ False ˤ롣
		* ǡ(ǧ)줿Ȥˤ True ˤ롣
		*/
	    public:
	       ITK_DEF_ACCESS(Bool & isSure(), { return isSure_; }) ;
	       Bool isSure_ ;

	       /*--------------------*/
	       /**
		*
		* ˡǡ(ǧ)줿Ȥ, isSure() 
		* True ˤ롣
		*/
	    public:
	       void letBeSure() { isSure() = True ; } ;

	       /*--------------------*/
	       /**
		*
		* ˡball 줿Ȥʤɤˡ
		* isSure()  False ˤ롣
		*/
	    public:
	       void letBeUnsure() { isSure() = False ; } ;

	       //------------------------------
	       // stamina
	       Flt _stamina;
	       Flt _effort;
	       Flt _recovery;

	       //------------------------------
	       // properties
	       Flt _dashpower_rate;
	       Flt _maxspeed;
	       Flt _decay;

	       SensorValue svalue ;
	       
	       //------------------------------
	       // FieldState::Object:: constructor
	       /*--------------------*/
	       /**
		*
		*/
	    public:
	       Object() { type = Other ; aux = 0 ; isSure() = False ; } ;

	       /*--------------------*/
	       /**
		*
		*/
	    public:
	       Object(const Object & src) {
		  copyFrom(src) ;
	       } ;

	       //------------------------------
	       // FieldState::Object:: copy
	       /*--------------------*/
	       /**
		*
		*/
	    public:
	       Object & operator= (const Object & src) {
		  return copyFrom(src) ;
	       } ;

	       /*--------------------*/
	       /**
		*
		*/
	    public:
	       Object & copyFrom(const Object & src) {
		  type = src.type ;
		  aux = src.aux ;
		  _posdir = src._posdir ;
		  posConf = src.posConf ;

		  /* <<< [Bun-0.3] <<< */
		  posVariant = src.posVariant ;
		  /* >>> [Bun-0.3] >>> */

		  _vel = src._vel ;
		  velConf = src.velConf ;
		  _time = src._time ;
		  
		  _stamina  = src._stamina;
		  _effort   = src._effort;
		  _recovery = src._recovery;
		  
		  _dashpower_rate = src._dashpower_rate;
		  _maxspeed       = src._maxspeed;
		  _decay          = src._decay;

		  svalue = src.svalue ;

		  isSure() = src.isSure() ;

		  return *this ;
	       } ;
	       
	       //------------------------------
	       // FieldState::Object:: access
	       /*--------------------*/
	       /**
		*
		*/
	    public:
	       ITK_DEF_ACCESS(CycleStep & time(), { return _time ; }) ;
	       
	       /*--------------------*/
	       /**
		*
		*/
	    public:
	       ITK_DEF_ACCESS(PosDir & posdir(), { return _posdir ; }) ;

	       /*--------------------*/
	       /**
		*
		*/
	    public:
	       ITK_DEF_ACCESS(Pos & pos(), { return _posdir.pos ; }) ;

	       /*--------------------*/
	       /**
		*
		*/
	    public:
	       ITK_DEF_ACCESS(Angle & hdir(), { return _posdir.hdir ; }) ;

	       /*--------------------*/
	       /**
		*
		*/
	    public:
	       ITK_DEF_ACCESS(Angle & bdir(), { return _posdir.bdir ; }) ;
	       
	       /*--------------------*/
	       /**
		*
		*/
	    public:
	       ITK_DEF_ACCESS(Vec & vel(), { return _vel ; }) ;

	       /*--------------------*/
	       /**
		*
		*/
	    public:
	       const Quality & quality() const {
		  return static_cast<const Quality>(static_cast<const int>(svalue.quality)) ; 
	       } ;

	       /*--------------------*/
	       /**
		*
		*/
	    public:
	       Quality & quality() {
		  return static_cast<Quality>(static_cast<int>(svalue.quality)) ; 
	       } ;

	       /*--------------------*/
	       /**
		*
		*/
	    public:
	       ITK_DEF_ACCESS(Flt & sdist(), { return svalue.dist ; }) ;

	       /*--------------------*/
	       /**
		*
		*/
	    public:
	       ITK_DEF_ACCESS(Flt & sdir(), { return svalue.dir ; }) ;

	       // by kshinoda 2002/03/20
	       ITK_DEF_ACCESS(Flt& stamina(),  { return _stamina;  });
	       ITK_DEF_ACCESS(Flt& effort(),   { return _effort;   }); 
	       ITK_DEF_ACCESS(Flt& recovery(), { return _recovery; });

	       // by kshinoda 2002/03/20
	       ITK_DEF_ACCESS(Flt& dashpower_rate(), { return _dashpower_rate; });
	       ITK_DEF_ACCESS(Flt& maxspeed(),       { return _maxspeed;       });
	       ITK_DEF_ACCESS(Flt& decay(),          { return _decay;          });


	       //------------------------------
	       // FieldState::Object:: isNewer
	       /*--------------------*/
	       /**
		*
		*/
	    public:
	       Bool isNewer(const CycleStep & from) const {
		  return time().count >= from.count ;
	       } ;

	       /*--------------------*/
	       /**
		*
		*/
	    public:
	       Bool isNewer(const UInt & from) const {
		  return time().count >= from ;
	       } ;
	       
	       /*--------------------*/
	       /**
		*
		*/
	    public:
	       Bool isNewer(const TimeVal & from) const {
		  return time().rtime >= from ;
	       } ;

	       //------------------------------
	       // FieldState::Object:: estimate
	       /*--------------------*/
	       /**
		*
		*/
	    public:
	       virtual Bool estimate(const SensorValue & oinfo,
				     const Object & origin,
				     const CycleStep & t) ;

	       //------------------------------
	       // FieldState::Object:: describe
	       /*--------------------*/
	       /**
		*
		*/
	    public:
	       virtual void describe(ostream & ostr, 
				     const Bool detailp = True) const {
		  ostr << "#Object[" << typestr(type) << ":" << aux 
		       << "]" ;

		  if(detailp) {
		     ostr << endl ;
		     ostr << "\t" << "  time=" << time() << endl ;
		     ostr << "\t" << "posdir=" << posdir() 
			  << " (conf=" << posConf << ")" << endl ;
		     ostr << "\t" << "   vel=" << vel() 
			  << " (conf=" << velConf << ")" << endl ;
		     ostr << "\t" << "  sure=" 
			  << (isSure() ? "yes" : "no") << endl ;
		     ostr << "\t" << "svalue=" ;
		     svalue.describe(ostr,detailp) ;
		  } 
	       } ;
	 } ;

	 //==================================================
	 // FieldState:: body part (FieldState)

	 //--------------------------------------------------
	 // FieldState:: member
      public:
	 Object _self ;
	 Angle _neckAngle ;

	 Bool _outofFieldP ;

	 Object _ball ;

	 Object _teammate[playerN] ;
	 Object _opponent[playerN] ;

	 UInt unknownN ;
	 Object _unknown[playerM] ;

#ifndef OLD_ESTIMATOR
	 PosEstimator posEstimator ;
#endif

	 //--------------------------------------------------
	 // FieldState:: constructor
	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 FieldState() {
	    init() ;
	 } ;

	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 FieldState(const FieldState & src) {
	    copyFrom(src) ;
	 } ;

	 //--------------------------------------------------
	 // FieldState:: init
	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 void init() {
	    neckAngle() = 0.0 ;
	    self().type = Object::Self ;
	    ball().type = Object::Ball ;
	    for(UInt i = 0 ; i < playerN ; i++) {
	       teammate(i).type = Object::Player ;
	       teammate(i).aux = i+1 ;
	       opponent(i).type = Object::Player ;
	       opponent(i).aux = i+1 ;
	    }
	    for(UInt i = 0 ; i < playerM ; i++) {
	       unknown(i).type = Object::Player ;
	       unknown(i).aux = Object::UnknownUNum ;
	    }
	 } ;

	 //--------------------------------------------------
	 // FieldState:: copy 
	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 FieldState & operator= (const FieldState & src) {
	    return copyFrom(src) ;
	 } ;

	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 FieldState & copyFrom(const FieldState & src) {
	    self() = src.self() ;
	    outofFieldP() = src.outofFieldP() ;
	    ball() = src.ball() ;
	    for(UInt i = 0 ; i < playerN ; i++) {
	       teammate(i) = src.teammate(i) ;
	       opponent(i) = src.opponent(i) ;
	    } ;

	    unknownN = src.unknownN ;
	    for(UInt i = 0 ; i < unknownN ; i++) {
	       unknown(i) = src.unknown(i) ; 
	    } ;
	    
	    return *this ;
	 } ;
	 
	 //--------------------------------------------------
	 // FieldState:: access
      public:

	 ITK_DEF_ACCESS(Object & self(), { return _self ; }) ;

	 ITK_DEF_ACCESS(CycleStep & time(), { return self().time() ; }) ;

	 ITK_DEF_ACCESS(PosDir & myposdir(),{ return self().posdir() ; }) ;
	 ITK_DEF_ACCESS(Flt & posConf(), { return self().posConf ;}) ;

	 /* <<< [Bun-0.3] <<< */
	 ITK_DEF_ACCESS(Flt & myposVar(), { return self().posVariant ;}) ;
	 /* >>> [Bun-0.3] >>> */

	 ITK_DEF_ACCESS(Pos & mypos(), { return myposdir().pos ; }) ;
	 ITK_DEF_ACCESS(Angle & mybdir(), { return myposdir().bdir ; }) ;
	 ITK_DEF_ACCESS(Angle & myhdir(), { return myposdir().hdir ; }) ;

	 
	 ITK_DEF_ACCESS(Vec & myvel(), { return self().vel() ; }) ;
	 ITK_DEF_ACCESS(Flt & velConf(), { return self().velConf ; }) ;

	 ITK_DEF_ACCESS(Bool & outofFieldP(), { return _outofFieldP ; }) ;
	 ITK_DEF_ACCESS(Angle & neckAngle(), { return _neckAngle ; }) ;

	 ITK_DEF_ACCESS(Object & ball(), { return _ball ; }) ;
	 ITK_DEF_ACCESS(Vec & ballpos(), { return ball().pos() ; }) ;
	 ITK_DEF_ACCESS(Vec & ballvel(), { return ball().vel() ; }) ;

	 /* <<< [Bun-0.3] <<< */
	 ITK_DEF_ACCESS(Flt & ballposVar(), { return ball().posVariant ; }) ;
	 /* >>> [Bun-0.3] >>> */
	 
	 ITK_DEF_ACCESS(Object & teammate(UInt i), { return _teammate[i] ; }) ;

	 ITK_DEF_ACCESS(Object & opponent(UInt i), { return _opponent[i] ; }) ;

	 ITK_DEF_ACCESS(Object & unknown(UInt i), { return _unknown[i] ; }) ;
	 
	 ITK_DEF_ACCESS(Object & player(UInt i), {
	    if(i<playerN) return teammate(i) ;
	    else          return teammate(i-playerN);
	 }) ;

	 //--------------------------------------------------
	 // FieldState:: estimate
	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 Bool estimate(BasePlayer & player) {
	    return estimate(player.sinfo.see,player) ;
	 } ;

	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 virtual Bool estimate(const SeeInfo & sinfo, BasePlayer & plyaer) ;

	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 virtual Bool estimateMyPos(const SeeInfo & sinfo, 
				    BasePlayer & player) ;

	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 virtual Bool estimateOther(const SeeInfo & sinfo, 
				    BasePlayer & player) ;

	 //--------------------------------------------------
	 // FieldState:: describe
	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 virtual void describe(ostream & ostr, 
			       const Bool detailp = True) const {
	    ostr << "#FieldState[" << (void *)this ;
	    ostr << "]" ;

	    if(detailp) {
	       ostr << endl ;

	       ostr << "#FieldState---------------------------------" << endl ;
	       ostr << "\t" << "myself=" ;
	       self().describe(ostr,detailp) ;
	       ostr << "   out of field?=" << outofFieldP() << endl ;
	       ostr << "#FieldState---------------------------------" << endl ;
	       ostr << "\t" << "ball=" ;
	       ball().describe(ostr,detailp) ;
	       ostr << "#FieldState---------------------------------" << endl ;
	       ostr << "\t" << "TEAMMATE" << endl ;;
	       for(UInt i = 0 ; i < playerN ; i++) {
		  ostr << "\t  teammate[" << i << "]:" ;
		  teammate(i).describe(ostr,detailp) ;
	       } ;
	       ostr << "#FieldState---------------------------------" << endl ;
	       ostr << "\t" << "OPPONENT" << endl ;
	       for(UInt i = 0 ; i < playerN ; i++) {
		  ostr << "\t<<<<<opponent[" << i << "]>>>>>:" ;
		  opponent(i).describe(ostr,detailp) ;
	       } 
	       ostr << "#FieldState---------------------------------" << endl ;
	       ostr << "\t" << "UNKNOWN" << endl ;
	       for(UInt i = 0 ; i < unknownN ; i++) {
		  ostr << "\t<<<<<unknown[" << i << "]>>>>>:" ;
		  unknown(i).describe(ostr,detailp) ;
	       } ;
	       ostr << "#FieldState=================================" << endl ;
	    } 
	 } ;
   } ;



} ;
//////////////////////////////////////////////////////////////////////
#endif



