// -*- Mode:C++ -*-
//Header:
//File: FullStateInfo.cc
//Title: Full State Info (implemented in version 7)
//Author: NODA, Itsuki
//Date: 2001/03/11
//

//ModifyHistory:
// 2001/03/11: Start to create this file
//EndModifyHistory:

#include "FullStateInfo.h"

namespace Soccer {

   //--------------------------------------------------
   // scan fullstate info

   Bool FullStateInfo::scan(SimpleSexp & s) {
      
      UInt len = s.length() ;
      if(len < 3) return False ;  /* 3 ??? */

      ///// time
      time.set(s.nth(1)->scanAsInt()) ;
      
      ///// other info

      playerN = 0 ;
      for(SimpleSexp * infolist = s.nthcdr(2) ; infolist->isCons() ;
	  infolist = infolist->cdr) {
	 SimpleSexp * info = infolist->car ;
	 SimpleSexp * label = info->car ;
	 SimpleSexp * args = info->cdr ;
	 
	 if      (label->equal(SC_WORDSTR(PMode))) {
	    playmode = scanAsScWord(args->nth(0)) ;
	 } 
	 else if (label->equal(SC_WORDSTR(VMode))) {
	    viewQuality = scanAsScWord(args->nth(0)) ;
	    viewQuality = scanAsScWord(args->nth(1)) ;
	 } 
	 else if (label->equal(SC_WORDSTR(Score))) {
	    scoreL = args->nth(0)->scanAsInt() ;
	    scoreR = args->nth(1)->scanAsInt() ;
	 } 
	 else if (label->equal(SC_WORDSTR(Ball))) {
	    ballpos.x = args->nth(0)->scanAsFlt() ;
	    ballpos.y = args->nth(1)->scanAsFlt() ;
	    ballvel.x = args->nth(2)->scanAsFlt() ;
	    ballvel.y = args->nth(3)->scanAsFlt() ;
	 }
	 else {  /* maybe player */
	    player[playerN].scan(info) ;
	    playerN++ ;
	 }
      }
      return True ;
   } ;

   //--------------------------------------------------
   // scan player info

   Bool FullStateInfo::Player::scan(SimpleSexp & s) {
      UInt len = s.length() ;
      if(len < 3) return False ; /* 3 ??? */
      
      name = scanAsScWord(s.nth(0)) ;
      if (isNull(name)) {
	 ITK_WRN("Illegal player's name in fullstate information."
		 << *(s.nth(0))) ;
	 return False ;
      } else if (SC_WID_L_1 <= name->id && name->id <= SC_WID_L_11) {
	 side = Left ;
	 unum = name->id - SC_WID_L_1 + 1 ;
      } else if (SC_WID_R_1 <= name->id && name->id <= SC_WID_R_11) {
	 side = Right ;
	 unum = name->id - SC_WID_R_1 + 1 ;
      } else {
	 ITK_WRN("Illegal player's name in fullstate information."
		 << *(s.nth(0))) ;
	 return False ;
      }

      posdir.pos.x = s.nth(1)->scanAsFlt();
      posdir.pos.y = s.nth(2)->scanAsFlt();

      vel.x = s.nth(3)->scanAsFlt();
      vel.y = s.nth(4)->scanAsFlt();
      
      posdir.bdir = s.nth(5)->scanAsFlt();
      posdir.hdir = s.nth(6)->scanAsFlt();

      stamina = s.nth(7)->scanAsFlt() ;
      effort = s.nth(8)->scanAsFlt() ;
      recovery = s.nth(9)->scanAsFlt() ;
      
      return True ;
   };

} ;
