// -*- Mode:C++ -*-
//Header:
//File: Hearnfo.h
//Title: HearInfo
//Author: NODA, Itsuki
//Date: 2001/03/05
//

//ModifyHistory:
// 2001/03/05: Start to create this file
//EndModifyHistory:

#ifndef _SPlayer_HearInfo_h_
#define _SPlayer_HearInfo_h_
////////////////////////////////////////////////////////////////////////

#include "itk/itk.h"
#include "sctypes.h"
#include "CycleStep.h"

namespace Soccer {

   using namespace Itk ;

   //======================================================================
   // HearInfo class

   /*--------------------*/
   /**
    *
    */

   class HearInfo : public WithDescriber {

	 //==================================================
	 // Sender Type
	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 enum SenderType {
	    Self = SC_WID_Self,
	    Referee = SC_WID_Referee,
	    Player = SC_WID_Player,
	    OnlineCoachL = SC_WID_OnlineCoachL,
	    OnlineCoachR = SC_WID_OnlineCoachR
	 } ;

	 /*--------------------*/
	 /**
	  *
	  */
#define SC_SENDERNAME_SELF      SC_WORDSTR(Self)

	 /*--------------------*/
	 /**
	  *
	  */
#define SC_SENDERNAME_REFEREE   SC_WORDSTR(Referee)

	 /*--------------------*/
	 /**
	  *
	  */
#define SC_SENDERNAME_PLAYER    SC_WORDSTR(Player)

	 /*--------------------*/
	 /**
	  *
	  */
#define SC_SENDERNAME_OLCOACHL	SC_WORDSTR(OnlineCoachL)

	 /*--------------------*/
	 /**
	  *
	  */
#define SC_SENDERNAME_OLCOACHR	SC_WORDSTR(OnlineCoachR)

	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 inline static const char* senderName(const SenderType& type) {
	    switch(type) {
	       case Self :        return SC_SENDERNAME_SELF ; 
	       case Referee :       return SC_SENDERNAME_REFEREE ; 
	       case Player :        return SC_SENDERNAME_PLAYER ;
	       case OnlineCoachL :  return SC_SENDERNAME_OLCOACHL ;
	       case OnlineCoachR :  return SC_SENDERNAME_OLCOACHR ;
	       default :            return SC_WORDSTR(Error) ;
	    }
	 } ;

	 //==================================================
	 // Main Body (HearInfo)

	 //--------------------------------------------------
	 // member
      public:
	 CycleStep time ;
	 SenderType sender ;
	 union {
	       Flt direction ;
	       ScWord* pmode ;
	 } aux ;

	 SubString messageCache ;

	 //--------------------------------------------------
	 // constructor
	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 HearInfo() {
	    messageCache.alloc(DfltBufSize) ;
	 } ;

	 //--------------------------------------------------
	 // copy
	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 HearInfo & copyFrom(const HearInfo & src) {
	    time = src.time ;
	    sender = src.sender ;
	    aux = src.aux ;
	    message().copy(src.message()) ;
	    
	    return *this ;
	 } ;	    

	 //--------------------------------------------------
	 // content
	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 ITK_DEF_ACCESS(SubString & message(), { return messageCache ; }) ;
	 
	 //--------------------------------------------------
	 // scan
	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 Bool scan(SimpleSexp & s) ;

	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 Bool scan(SimpleSexp * s) {
	    return scan(*s) ;
	 } ;

	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 Bool scan(Buffer & b) { 
	    return scan(SimpleSexp::scan(b)) ;
	 } ;

	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 Bool scan(SubString & str) {
	    Buffer b(str) ;
	    return scan(b) ;
	 } ;

	 //--------------------------------------------------
	 // describe
      public:
	 virtual void describe(ostream & ostr, 
			       const Bool detailp = True) const {
	    ostr << "#HearInfo[t=" << time << ", "
		 << "sender=" << senderName(sender) << "]" ;
	    if(detailp) {
	       ostr << endl ;
	       ostr << "\tmessage=\"" << message() << "\"" << endl ;
	    }
	 } ;
   } ;


} ;
////////////////////////////////////////////////////////////////////////
#endif
