// -*- Mode:C++ -*-
//Header:
//File: Newton2dKFilter.h
//Title: Bun Player for chagamma
//Author: NODA, Itsuki
//Date: 2002/04/10
//

//ModifyHistory:
// 2001/03/02: Start to create this file
// 2002/04/10: Branch from chagama.cc
//EndModifyHistory:

#ifndef _Bun_Newton2dKFilter_h_
#define _Bun_Newton2dKFilter_h_
////////////////////////////////////////////////////////////////////////

#include "itk/itk.h"

namespace Soccer {

   using namespace Itk ;

   typedef Flt Val ;

   class ValMatrix ;
   class ValVector ;

   //======================================================================
   // class ValVector
   /*--------------------*/
   /**
    *
    */

   class ValVector : public WithDescriber {
      public:
	 virtual const char* className() const { return "V" ; } ;
	    
	 //--------------------------------------------------
	 //	:: ValVector
	 // members & access
	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 ITK_DEF_ACCESS(Int & size(), { return size_ ; }) ;
	 Int size_ ;

	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 ITK_DEF_ACCESS(Val & val(const Int & i), { return val()[i];});
	 ITK_DEF_ACCESS3(Val *, & val(), { return val_ ;}) ;
	 Val * val_ ;

	 //--------------------------------------------------
	 //	:: ValVector
	 // init and allocation
	 /*--------------------*/
	 /**
	  * constructor
	  */
      public:
	 ValVector() { alloc(0); } ;
	 ValVector(const Int & s) { alloc(s) ;} ;

	 /*--------------------*/
	 /**
	  * allocation
	  */
      public:
	 void 
	 alloc(Int s = 0) {
	    size() = s ;
	    if(size() == 0) {
	       val() = ITK_NULLPTR ;
	    } else {
	       val() = new Val[size()] ;
	    }
	 } ;

	 //--------------------------------------------------
	 //	:: ValVector
	 // operators
	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 ValVector & 
	 operator=(const ValVector & src) {
	    return set(src) ;
	 } ;

	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 ValVector & operator+=(const ValVector & src) {
	    return inc(src) ;
	 } ;

	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 ValVector & operator-=(const ValVector & src) {
	    return dec(src) ;
	 } ;

	 //--------------------------------------------------
	 //	:: ValVector
	 // set values
	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 ValVector & 
	 set(const Val & v) {
	    for(Int i = 0 ; i < size() ; i++) val(i) = v ;
	    return *this ;
	 } ;

	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 ValVector & 
	 set(const ValVector & src) {
	    for(Int i = 0 ; i < size() ; i++) val(i) = src.val(i) ;
	    return *this ;
	 } ;

	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 ValVector & 
	 setMul(const ValMatrix & w, 
		const ValVector & v) {
	    return incMul(w,v,True) ;
	 } ;

	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 ValVector & 
	 setZero() { return set(0.0);} ;

	 //--------------------------------------------------
	 //	:: ValVector
	 // inc / dec
	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 ValVector & 
	 inc(const ValVector & src) {
	    for(Int i = 0 ; i < size() ; i++) val(i) += src.val(i) ;
	    return *this ;
	 } ;

	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 ValVector & 
	 incMul(const ValMatrix & w, 
		const ValVector & v,
		const Bool clearp = False) ;


	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 ValVector & 
	 dec(const ValVector & src) {
	    for(Int i = 0 ; i < size() ; i++) val(i) -= src.val(i) ;
	    return *this ;
	 } ;

	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 ValVector & 
	 decMul(const ValMatrix & w, 
		const ValVector & v,
		const Bool clearp = False) ;


	 //--------------------------------------------------
	 //	:: ValVector
	 // output & describe
	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 void showElements(ostream & ostr) const {
	    for(Int i = 0 ; i < size() ; i++) {
	       ostr << setiosflags(ios::fixed) << setfill(' ')
		    << setw(10) << setprecision(6) 
		    << val(i) << " " ;
	    }
	 } ;

	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 virtual void describeShortBody(ostream & ostr) const {
	    ostr << size() << ":" << Ptr(this) ;
	 } ;

	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 virtual void describeDetail(ostream & ostr) const {
	    ostr << endl << "\t" ;
	    ostr << "val=" ;
	    showElements(ostr) ;
	    ostr << endl ;
	 } ;
   } ;

   //======================================================================
   // class ValMatrix
   /*--------------------*/
   /**
    *
    */

   class ValMatrix : public WithDescriber {
      public:
	 virtual const char* className() const { return "M" ; } ;

	 //--------------------------------------------------
	 //	:: VamMatrix
	 // members & access
	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 ITK_DEF_ACCESS(Int & sizeN(), { return sizeN_ ; }) ;
	 Int sizeN_ ;

	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 ITK_DEF_ACCESS(Int & sizeM(), { return sizeM_ ; }) ;
	 Int sizeM_ ;

	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 ITK_DEF_ACCESS(Val & val(const Int & i, const Int & j), {
	    return val(i)[j] ;
	 }) ;
	 ITK_DEF_ACCESS3(Val *, & val(const Int & i), { 
	    return val()[i];});
	 ITK_DEF_ACCESS3(Val **, & val(), { return val_ ;}) ;
	 Val ** val_ ;

	 //--------------------------------------------------
	 //	:: VamMatrix
	 // init and allocation
	 /*--------------------*/
	 /**
	  * constructor
	  */
      public:
	 ValMatrix() { alloc(0,0) ; } ;
	 ValMatrix(const Int n, const Int m) { alloc(n,m) ; } ;

	 /*--------------------*/
	 /**
	  * allocation
	  */
      public:
	 void 
	 alloc(Int n = 0, Int m = 0) {
	    sizeN() = n ;
	    sizeM() = m ;
	    if(sizeN() == 0) {
	       val() = ITK_NULLPTR ;
	    } else {
	       val() = new Val * [sizeN()] ;
	       for(Int i = 0 ; i < sizeN() ; i++) {
		  val(i) = new Val[sizeM()] ;
	       }
	    }
	 } ;

	 //--------------------------------------------------
	 //	:: VamMatrix
	 // operators
	 /*--------------------*/
	 /**
	  * 
	  */
      public:
	 ValMatrix & 
	 operator=(const ValMatrix & src) {
	    return set(src) ;
	 } ;

	 /*--------------------*/
	 /**
	  * 
	  */
      public:
	 ValMatrix & 
	 operator+=(const ValMatrix & src) {
	    return inc(src) ;
	 } ;

	 /*--------------------*/
	 /**
	  * 
	  */
      public:
	 ValMatrix & 
	 operator-=(const ValMatrix & src) {
	    return dec(src) ;
	 } ;

	 //--------------------------------------------------
	 //	:: VamMatrix
	 // set values
	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 ValMatrix & 
	 set(const Val & v) {
	    for(Int i = 0 ; i < sizeN() ; i++) {
	       for(Int j = 0 ; j < sizeM() ; j++) {
		  val(i,j) = v ;
	       } 
	    }
	    return *this ;
	 } ;

	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 ValMatrix & 
	 set(const ValMatrix & v) {
	    for(Int i = 0 ; i < sizeN() ; i++) {
	       for(Int j = 0 ; j < sizeM() ; j++) {
		  val(i,j) = v.val(i,j) ;
	       } 
	    }
	    return *this ;
	 } ;

	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 ValMatrix & 
	 setZero() { return set(0.0) ; } ;

	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 ValMatrix & 
	 setDiag(const Val & v, Bool cleanp = True) {
	    if(cleanp) setZero() ;
	    for(Int i = 0 ; i < sizeN() ; i++) val(i,i) = v ;
	    return *this ;
	 } ;

	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 ValMatrix & 
	 setTrans(const ValMatrix & v) {
	    for(Int i = 0 ; i < sizeN() ; i++) {
	       for(Int j = 0 ; j < sizeM() ; j++) {
		  val(i,j) = v.val(j,i) ;
	       } 
	    }
	    return *this ;
	 } ;

	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 ValMatrix & 
	 setMul(const ValMatrix & w1,
		const ValMatrix & w2) {
	    return incMul(w1,w2,True) ;
	 } ;

	 //--------------------------------------------------
	 //	:: VamMatrix
	 // inc / dec
	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 ValMatrix & 
	 inc(const ValMatrix & src) {
	    for(Int i = 0 ; i < sizeN() ; i++) {
	       for(Int j = 0 ; j < sizeM() ; j++) {
		  val(i,j) += src.val(i,j) ;
	       } ;
	    } ;
	    return *this ;
	 } ;

	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 ValMatrix & 
	 dec(const ValMatrix & src) {
	    for(Int i = 0 ; i < sizeN() ; i++) {
	       for(Int j = 0 ; j < sizeM() ; j++) {
		  val(i,j) -= src.val(i,j) ;
	       } ;
	    } ;
	    return *this ;
	 } ;

	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 ValMatrix & 
	 incMul(const ValMatrix & w1,
		const ValMatrix & w2,
		const Bool clearp = False) ;

	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 ValMatrix & 
	 decMul(const ValMatrix & w1,
		const ValMatrix & w2,
		const Bool clearp = False) ;

	 //--------------------------------------------------
	 //	:: VamMatrix
	 // output & describe
	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 void 
	 showElements(ostream & ostr) const {
	    for(Int i = 0 ; i < sizeN() ; i++) {
	       ostr << endl << "\t" ;
	       for(Int j = 0 ; j < sizeM() ; j++) {
		  ostr << setiosflags(ios::fixed) << setfill(' ')
		       << setw(10) << setprecision(6) 
		       << val(i,j) << " " ;
	       }
	    }
	 } ;

	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 virtual void describeShortBody(ostream & ostr) const {
	    ostr << sizeN() << "," << sizeM() << ":" << Ptr(this) ;
	 } ;

	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 virtual void describeDetail(ostream & ostr) const {
	    ostr << endl << "    " ;
	    ostr << "val=" ;
	    showElements(ostr) ;
	    ostr << endl ;
	 } ;
   } ;

   //======================================================================
   // class Newton2dKFilter
   /*--------------------*/
   /**
    * 
    * Kalman filter class for Newtonian 2 dimentional world.
    * 
    * {\bf :} 
    *
    * ưǥϰʲɽȤ롣
    * 	\[ 
    * 		x_{true}^{t+1} = F x_{true}^{t} + q 
    * 	\]
    * , $q$ ϸ, 
    * μȤäͽ $x_f$ Ȥ롣ʤ,
    * 	\[ 
    * 	        x_f^{t+1} = F x_f^{t} 
    * 	\]
    * λ, ͽʬ $P_f$ ϡ
    * 	\[ 
    * 		P_f^{t+1} = F P_f^{t} F^T + Q 
    *	\]
    * Ѳ롣
    * , $Q$  $q$ ζʬ
    * 
    * ¬ $x_o$ 줿Ȥˤϡ
    * ʲΤ褦˥ǡ礷ƿͤ롣
    *
    * , $x_o$ ϴ¬ $H$ ѤưʲΤ褦˵ޤΤȤ롣
    * 	\[
    *		x_o^t = H x_{true}^t + r
    *	\]
    * ޤ$R$  $r$ ζʬȤ롣
    * Ѥ,
    * ޤޥ󥲥 $K$ 롣
    *	\[
    *		K = P_f H^T (H P_f H^T + R)^{-1}
    *	\]
    * Ѥ,ʲΤ褦˿ $x_a$ 롣
    * 	\[
    *		x_a = x_f + K (x_o - H x_f)
    *	\]
    * λ, ʬ $P_a$ ϡ
    * 	\[
    *		P_a = P_f - K H P_f = (I - K H) P_f 
    *		    = (H^T R^{-1} + P_f^{-1})^{-1}
    *	\]
    *  $x_a$, $P_a$ 򿷤 $x_f$, $P_f$ ȤơͽԤ
    *
    * {\bf 2˥塼ȥưˤ:}
    *
    * ѿ $x$ , 
    * ֤ x-y, ®٤ x-y, ®٤ x-y 6ΤȤ롣
    * λ, $F$ ϰʲΤ褦ʹȤ롣
    * 	\[
    *		F = \begin{array}{|cccccc|}
    *		  1 &   & 1 &   & 1 &   \\
    *		    & 1 &   & 1 &   & 1 \\
    *		    &   & \alpha &  & \alpha & \\
    *		    &   &   & \alpha &  & \alpha \\
    *		  0 & 0 & 0 & 0 & 0 & 0 \\
    *		  0 & 0 & 0 & 0 & 0 & 0 
    *		\end{array}
    *	\]
    * ޤ¬ $H$ ϡ֤¬ǤΤȤ롣
    * 	\[
    *		H = \begin{array}{|cccccc|}
    *		  1 &   &   &   &   &   \\
    *		    & 1 &   &   &   &   \\
    *		    &   &   &   &   &   \\
    *		    &   &   &   &   &   \\
    *		    &   &   &   &   &   \\
    *		    &   &   &   &   &  
    *		\end{array}
    *	\]
    * ޤƸʬ, гʬΤߤȤ롣
    * 
    * {\bf Ȥ:} 
    *
    * \begin{itemize}
    *  \item ͽ¬
    *   \begin{enumerate}
    *	 \item kfilter.vXf() ˸ͤ
    *	 \item kfilter.wPf() ˸ͤ
    *	 \item kfilter.estCycle() ƤӽФ
    *	 \item kfilter.wPf() Ǽ
    *	 \item kfilter.vXf() ѿͤФ
    *   \end{enumerate}
    *
    *  \item ¬
    *   \begin{enumerate}
    *	 \item kfilter.vXf() ˸ͤ
    *	 \item kfilter.wPf() ˸ͤ
    *    \item kfilter.setParamObserv() Ǵ¬ʬꡣ
    *	 \item kfilter.mergeSensor(pos\_x,pos\_y) ƤӽФ
    *	 \item kfilter.wPf() Ǽ
    *	 \item kfilter.vXf() ѿͤФ
    *   \end{enumerate}
    * \end{itemize}
    * 
    */

   class Newton2dKFilter : public WithDescriber {
      public:
	 virtual const char* className() const { return "Newton2dKFilter" ; } ;

	 //--------------------------------------------------
	 //	:: Newton2dKFilter
	 // member and access

	 //----------------------------------------
	 //	:: Newton2dKFilter
	 // parameters
	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 ITK_DEF_ACCESS(Val & alpha(), { return alpha_;}) ;
	 Val alpha_ ;

	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 ITK_DEF_ACCESS(Val & sigmaPxx(), { return sigmaPxx_;}) ;
	 Val sigmaPxx_ ;

	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 ITK_DEF_ACCESS(Val & sigmaPyy(), { return sigmaPyy_;}) ;
	 Val sigmaPyy_ ;

	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 ITK_DEF_ACCESS(Val & sigmaVxx(), { return sigmaVxx_;}) ;
	 Val sigmaVxx_ ;

	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 ITK_DEF_ACCESS(Val & sigmaVyy(), { return sigmaVyy_;}) ;
	 Val sigmaVyy_ ;

	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 ITK_DEF_ACCESS(Val & sigmaAxx(), { return sigmaAxx_;}) ;
	 Val sigmaAxx_ ;

	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 ITK_DEF_ACCESS(Val & sigmaAyy(), { return sigmaAyy_;}) ;
	 Val sigmaAyy_ ;

	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 ITK_DEF_ACCESS(Val & sigmaOPxx(), { return sigmaOPxx_;}) ;
	 Val sigmaOPxx_ ;

	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 ITK_DEF_ACCESS(Val & sigmaOPyy(), { return sigmaOPyy_;}) ;
	 Val sigmaOPyy_ ;

	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 ITK_DEF_ACCESS(Int & size(), { return size_ ;}) ;
	 Int size_ ;

	 //----------------------------------------
	 //	:: Newton2dKFilter
	 // matrixes
	 /*--------------------*/
	 /**
	  *
	  * ϳطϤ򵭽Ҥ
	  */
      public:
	 ITK_DEF_ACCESS(Val & wF(const Int & i,
				 const Int & j), { 
	    return wF().val(i,j) ; }) ;
	 ITK_DEF_ACCESS(ValMatrix & wF(), { return wF_ ;}) ;
	 ValMatrix wF_ ;

	 /*--------------------*/
	 /**
	  *
	  * wFt оι
	  */
      public:
	 ITK_DEF_ACCESS(Val & wFt(const Int & i,
				  const Int & j), { 
	    return wFt().val(i,j) ; }) ;
	 ITK_DEF_ACCESS(ValMatrix & wFt(), { return wFt_ ;}) ;
	 ValMatrix wFt_ ;

	 /*--------------------*/
	 /**
	  *
	  * ͽʬ
	  */
      public:
	 ITK_DEF_ACCESS(Val & wQ(const Int & i,
				 const Int & j), { 
	    return wQ().val(i,j) ; }) ;
	 ITK_DEF_ACCESS(ValMatrix & wQ(), { return wQ_ ;}) ;
	 ValMatrix wQ_ ;

	 /*--------------------*/
	 /**
	  *
	  * ͽ͸ʬ
	  */
      public:
	 ITK_DEF_ACCESS(Val & wPf(const Int & i,
				  const Int & j), { 
	    return wPf().val(i,j) ; }) ;
	 ITK_DEF_ACCESS(ValMatrix & wPf(), { return wPf_ ;}) ;
	 ValMatrix wPf_ ;

	 /*--------------------*/
	 /**
	  *
	  * ץ󥰥ե륿
	  */
      public:
	 ITK_DEF_ACCESS(Val & wH(const Int & i,
				 const Int & j), { 
	    return wH().val(i,j) ; }) ;
	 ITK_DEF_ACCESS(ValMatrix & wH(), { return wH_ ;}) ;
	 ValMatrix wH_ ;

	 /*--------------------*/
	 /**
	  *
	  * 󥵡ʬ
	  */
      public:
	 ITK_DEF_ACCESS(Val & wR(const Int & i,
				 const Int & j), { 
	    return wR().val(i,j) ; }) ;
	 ITK_DEF_ACCESS(ValMatrix & wR(), { return wR_ ;}) ;
	 ValMatrix wR_ ;

	 /*--------------------*/
	 /**
	  *
	  * ޥ󥲥
	  */
      public:
	 ITK_DEF_ACCESS(Val & wK(const Int & i,
				 const Int & j), { 
	    return wK().val(i,j) ; }) ;
	 ITK_DEF_ACCESS(ValMatrix & wK(), { return wK_ ;}) ;
	 ValMatrix wK_ ;

	 /*--------------------*/
	 /**
	  *
	  * ͸ʬ
	  */
      public:
	 ITK_DEF_ACCESS(Val & wPa(const Int & i,
				  const Int & j), { 
	    return wPa().val(i,j) ; }) ;
	 ITK_DEF_ACCESS(ValMatrix & wPa(), { return wPa_ ;}) ;
	 ValMatrix wPa_ ;


	 //----------------------------------------
	 //	:: Newton2dKFilter
	 // vectors
	 /*--------------------*/
	 /**
	  *
	  * ͽ
	  */
      public:
	 ITK_DEF_ACCESS(Val & vXf(const Int & i), { 
	    return vXf().val(i) ; }) ;
	 ITK_DEF_ACCESS(ValVector & vXf(), { return vXf_ ;}) ;
	 ValVector vXf_ ;

	 /*--------------------*/
	 /**
	  *
	  * ƥȤѤ뿿
	  */
      public:
	 ITK_DEF_ACCESS(Val & vXt(const Int & i), { 
	    return vXt().val(i) ; }) ;
	 ITK_DEF_ACCESS(ValVector & vXt(), { return vXt_ ;}) ;
	 ValVector vXt_ ;

	 /*--------------------*/
	 /**
	  *
	  * 󥵡
	  */
      public:
	 ITK_DEF_ACCESS(Val & vXo(const Int & i), { 
	    return vXo().val(i) ; }) ;
	 ITK_DEF_ACCESS(ValVector & vXo(), { return vXo_ ;}) ;
	 ValVector vXo_ ;

	 /*--------------------*/
	 /**
	  * 
	  * ͽͤȥ󥵡ͤ줿
	  */
      public:
	 ITK_DEF_ACCESS(Val & vXa(const Int & i), { 
	    return vXa().val(i) ; }) ;
	 ITK_DEF_ACCESS(ValVector & vXa(), { return vXa_ ;}) ;
	 ValVector vXa_ ;


	 //----------------------------------------
	 //	:: Newton2dKFilter
	 // work area
	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 ITK_DEF_ACCESS(Val & wW0(const Int & i,
				  const Int & j), { 
	    return wW0().val(i,j) ; }) ;
	 ITK_DEF_ACCESS(ValMatrix & wW0(), { return wW0_ ;}) ;
	 ValMatrix wW0_ ;

	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 ITK_DEF_ACCESS(Val& wW1(const Int & i,
				 const Int & j), { 
	    return wW1().val(i,j) ; }) ;
	 ITK_DEF_ACCESS(ValMatrix & wW1(), { return wW1_ ;}) ;
	 ValMatrix wW1_ ;


	 /*--------------------*/
	 /**
	  *
	  */
      public:
	 ITK_DEF_ACCESS(Val & vXw(const Int & i), { 
	    return vXw().val(i) ; }) ;
	 ITK_DEF_ACCESS(ValVector & vXw(), { return vXw_ ;}) ;
	 ValVector vXw_ ;

	 //--------------------------------------------------
	 //	:: Newton2dKFilter
	 // initialize

      public:
	 static const Flt dfltAlpha = 0.9 ;
	 static const Flt dfltSigmaPos = 0.01 ;
	 static const Flt dfltSigmaVel = 0.01 ;
	 static const Flt dfltSigmaAcc = 0.5 ;
	 static const Flt dfltSigmaObserv = 0.3 ;

	 static const Int dfltSize = 6 ;

	 /*--------------------*/
	 /**
	  * 
	  */
      public:
	 void init() ;

	 /*--------------------*/
	 /**
	  * ѥ᡼
	  *
	  * ѥ᡼νͤǥեȤꡣ
	  */
	 void initParam() ;

	 /*--------------------*/
	 /**
	  * ѥ᡼
	  *
	  * ѥ᡼νͤǥեȤꡣ
	  */
	 void setParam(Val a = dfltAlpha,
		       Val sigmaPos = dfltSigmaPos,
		       Val sigmaVel = dfltSigmaVel,
		       Val sigmaAcc = dfltSigmaAcc,
		       Val sigmaObserv = dfltSigmaObserv,
		       Int s = dfltSize) ;

	 /* <<< [Bun-0.3] <<< */

	 /*--------------------*/
	 /**
	  * ͽ¬Υѥ᡼
	  *
	  * ͽ¬Υʬꡣ
	  * setMatrixP = True ξ, ޥȥꥯˤͤ򥻥åȡ
	  */
	 void setParamPredNoise(Val sigmaPos, Val sigmaVel, Val sigmaAcc, 
				Bool setMatrixP = True) ;

	 /*--------------------*/
	 /**
	  * ¬ѥ᡼
	  *
	  * ¬Υʬꡣ
	  * setMatrixP = True ξ, ޥȥꥯˤͤ򥻥åȡ
	  */
	 void setParamObserv(Val sigmaObserv, Bool setMatrixP = True) ;

	 /* >>> [Bun-0.3] >>> */

	 /*--------------------*/
	 /**
	  * ޥȥꥯ
	  *
	  * ѥ᡼˽äƥޥȥꥯνͤ
	  */
	 void initMatrix(Bool allocp = False) ;

	 /*--------------------*/
	 /**
	  * ޥȥꥯꥢ
	  */
	 void allocMatrix() ;

	 //--------------------------------------------------
	 //	:: Newton2dKFilter
	 // initialize
	 /*--------------------*/
	 /**
	  *
	  * ˴ŤѲͽ¬ȳƼѥ᡼Ԥ
	  */
      public:
	 void estCycle() ;

	 /*--------------------*/
	 /**
	  * 
	  * 󥵡礹롣
	  */
      public:
	 void mergeSensor(Flt px, Flt py) ;

	 //--------------------------------------------------
	 //	:: Newton2dKFilter
	 // for test
	 /*--------------------*/
	 /**
	  *
	  * ƥȤΤγΥߥ졼
	  */
      public:
	 void testEnvCycle() ;

	 /*--------------------*/
	 /**
	  *
	  * testEnvCycle() Ȥä, Kalman Filter ΥƥȤ򤹤1
	  */
      public:
	 Bool testCycle() ;

	 /*--------------------*/
	 /**
	  *
	  * testCycle() ƤӽФ
	  */
      public:
	 void testRun() ;
	 
   } ;
} ;



////////////////////////////////////////////////////////////////////////
#endif

