// -*- Mode:C++ -*-
//Header:
//File: SimplePlayer.cc
//Title: Base Player for Soccer Server 7
//Author: NODA, Itsuki
//Date: 2001/03/10
//

//ModifyHistory:
// 2001/03/10: Start to create this file
//EndModifyHistory:

#include "SimplePlayer.h"

namespace Soccer {

   const Flt SimplePlayer::TurnCutOffAngle ;

   //======================================================================
   // state manipuration

   FieldState & SimplePlayer::pushState(const CycleStep & time, 
					const FieldState & s) {
      FieldState & fs = stateEntry(time.count) ;
      fs.copyFrom(s) ;
      senseStateTime = time ;
      estimateStateTime = time ;

      /* <<< [I.Noda:2001.07.26] <<< */
      worldmodel.mergeSensedData(time.count,s) ;
      /* >>> [I.Noda:2001.07.26] >>> */

      return fs ;
   }

   //======================================================================
   // analyze see info

   Bool SimplePlayer::analyzeSensorSee() { 
      currentState().estimate(*this) ;
      pushState(vtime(),currentState()) ;
      if(stime() <= vtime()) { stime() = vtime() ; } ;
      return True ; 
   } ;

   //======================================================================
   // analyze hear info

   Bool SimplePlayer::analyzeSensorHear() { 
      if(sinfo.hear.sender == HearInfo::Referee) {
	 cvPlayMode.lock() ; {
	    playmode() = sinfo.hear.aux.pmode ;
	    cvPlayMode.broadcast() ;
	    //ITK_DBG(*playmode()) ;
	 } ; cvPlayMode.unlock() ;
      }
      atime() = sinfo.time ;
      if(stime() <= atime()) { stime() = atime() ; } ;
      return True ;
   } ;

   //======================================================================
   // analyze sensebody info

   Bool SimplePlayer::analyzeSensorSenseBody() { 

      myneck() = sinfo.senseBody.headAngle ;
      btime() = sinfo.time ;
      stime() = btime() ;

      currentState().self().stamina() = sinfo.senseBody.stamina;
      currentState().self().effort()  = sinfo.senseBody.effort;
      currentState().self().recovery() = 0;
      
      worldmodel.worktime = stime().count ;
      WorldModel::State & state = 
	 worldmodel.estState(stime().count,True) ;
      state.mergeMyspeed(sinfo.senseBody.speed) ;
      worldmodel.action(stime().count).setup(stime().count) ;

      return True ;
   } ;

   //======================================================================
   // analyze see info

   Bool SimplePlayer::analyzeSensorFullState() { 

      taughtState().time() = sinfo.fullstate.time ;
      taughtState().ballpos() = sinfo.fullstate.ballpos ;
      taughtState().ballvel() = sinfo.fullstate.ballvel ;
      FullStateInfo::Player * p = ITK_NULLPTR;
      for(UInt i = 0 ; i < sinfo.fullstate.playerN ; i++) {
	 if(sinfo.fullstate.player[i].side == side &&
	    sinfo.fullstate.player[i].unum == unum) {
	    p = &(sinfo.fullstate.player[i]) ;
	    break;
	 }
      }
      taughtState().myposdir() = p->posdir ;
      taughtState().myvel() = p->vel ;
      
      fstime() = sinfo.time ;
      stime() = fstime() ;
      return True ;
   } ;

   //======================================================================
   // Command Timing

   Bool SimplePlayer::waitNextBestComTime(UInt marginMS) {

      /* <<< [2002/05/07:I.Noda] <<< */
      /*
       *  because sleep will delay 10ms in maximum, so that the time
       *  marginMS should be increase 10ms.
       */
      marginMS += 10 ;
      /* >>> [2002/05/07:I.Noda] >>> */
      
      if(bctime() >= btime()) cvSenseBody.lockWait() ;
      TimeVal timeup = btime().rtime + TimeVal(0,100-marginMS) ;
      //ITK_VBS(1,"wait :" << timeup) ;

      /* <<< [Bun-0.2] <<< */
      /* using sleepBusy instead of sleep 
       * !! This does not work.
       */
      // sleepUntil(timeup) ;
      /* === [Bun-0.2] === */
      BasePlayer::sleepUntil(timeup) ;
      /* >>> [Bun-0.2] >>> */

      return True ;
   } 

   //======================================================================
   // Wrapper for Super Class
  
   /* <<< [Bun-0.2] <<< */
   /* 
    * body-command sending time  sense-body time 
    * 
    */
   Bool SimplePlayer::sendCommandPre(Bool bodyp) {
      Bool r = ThreadedPlayer::sendCommandPre(bodyp) ;

      /* <<< [Bun-0.4] <<< */
      //nextSendBodyTime = btime().rtime + TimeVal(0,SendBodyTic,0) ;
      /* === [Bun-0.4] === */
      /* body command λΤߡꥻåȤԤʤȤʤ
       */
      if(bodyp) {
	 TimeVal diff = nextSendBodyTime - btime().rtime ;
	 if(diff > TimeVal(1,0,0)) { // 1ðʾ٤Ϥʤˤꤢ
	    ITK_WRN("Something Wrong in sending body command." 
		    << ITK_D(btime()) << ITK_D(nextSendBodyTime)) ;
	 } else {
	    nextSendBodyTime = btime().rtime + TimeVal(0,SendBodyTic,0) ;
	 }
      }
      /* >>> [Bun-0.4] >>> */
      
      return r ;
   } ;
   /* >>> [Bun-0.2] >>> */

   Bool SimplePlayer::sendCommandPost(Bool bodyp) {
      ctime().set(stime().count,True) ;
      if(bodyp) bctime().set(stime().count,True) ;
      return ThreadedPlayer::sendCommandPost(bodyp) ;
   } ;

   //======================================================================
   // Basic Command

   //--------------------------------------------------
   // move

   Bool SimplePlayer::move(const Flt & x, const Flt & y) { 
      return sendMove(x,y) ; 
   } ;

   //--------------------------------------------------
   // turn

   Bool SimplePlayer::turn(const Flt & power_, const Flt & cutoff) { 
      Flt power = normalizeAngle(power_) ;
      if(Abs(power) < cutoff) return False ;

      Flt inertia_moment = 5.0 ; /* ***** from server.conf */
      Flt factor = (1.0 + inertia_moment * myspeed()) ;
      Flt realpower = power * factor ;

      Bool r = False ;

      /* <<< [Bun-0.2] <<< */

      Flt restdir = 0.0 ;

      /* If the realpower is out-of-range for turn command, then
       * it separate multiple turn commands */

      if(Abs(realpower) > SC_MAXMOMENT) {
	 if(realpower > 0.0) {
	    restdir = realpower - SC_MAXMOMENT ;
	    realpower = SC_MAXMOMENT ;
	 } else {
	    restdir = realpower + SC_MAXMOMENT ;
	    realpower = -SC_MAXMOMENT ;
	 }
	 restdir /= factor ;
      }
      

      r = sendTurn(realpower) ;

      if(r) {
	 mybdir() += power ;
	 myhdir() += power ;

	 worldmodel.action(stime().count).setTurn(realpower) ;

	 if(restdir != 0.0) {
	    waitNextBestComTime(0) ;
	    r = SimplePlayer::turn(restdir) ;
	 }
      }

      /* >>> [Bun-0.2] >>> */
      
      
      return r ;
   } ;

   //--------------------------------------------------
   // dash

   Bool SimplePlayer::dash(const Flt & power) { 
      Bool r = sendDash(power) ; 

      if(r) {
	 worldmodel.action(stime().count).setDash(power) ;
      }

      return r ;
   } ;

   //--------------------------------------------------
   // kick

   Bool SimplePlayer::kick(const Flt & power, const Flt & dir_) {

     // KickPowerRate ν
     /*
     WorldModel::State cstate = worldmodel.estState(stime().count) ;
     Vec bpos = cstate.ball().pos();
     Vec spos = cstate.self().pos();
     Angle sbangle = cstate.self().bdir();
     Flt tmpdist = spos.distance(bpos) - SC_PLAYER_SIZE - SC_BALL_SIZE;
     Flt tmpdir  = fabs(normalizeAngle((spos - bpos).angle() - sbangle));
     Flt kickpowerrate = 
       SC_KICK_POWER_RATE * (1.0 - 0.25*(tmpdist/SC_KICKABLE_MARGIN)
			     - 0.25 * (tmpdir/SC_MAXMOMENT));
     Flt kickpower = power;
     if(kickpowerrate != 0)
       kickpower /= kickpowerrate;

     if(kickpower > SC_MAXPOWER){
       kickpower = SC_MAXPOWER;
     } else {
       kickpower = 0;
     }
     */
      Flt dir = normalizeAngle(dir_) ;

      Bool r = sendKick(power,dir) ;

      if(r) {
	 worldmodel.action(stime().count).setKick(power,dir) ;
      }

      return r ;
   } ;

   //--------------------------------------------------
   // goalie catch

   Bool SimplePlayer::gcatch(const Flt & dir) {
      return sendCatch(dir) ;
   } ;

   //--------------------------------------------------
   // turn neck

   Bool SimplePlayer::head(const Flt & dir) {
      Flt neck = normalizeAngle(myneck() + dir) ;
      if(neck >  90.0) neck =  90.0 ;
      if(neck < -90.0) neck = -90.0 ;

      Flt moment = normalizeAngle(neck - myneck()) ;

      Bool r = sendTurnNeck(moment) ;
      
      if(r) {
	 myneck() = neck ;
	 myhdir() = mybdir() + myneck() ;

	 worldmodel.action(stime().count).setHead(moment) ;
      }

      return r ;
   } ;

   //--------------------------------------------------
   // say

   Bool SimplePlayer::say(const SubString & msg) {
      return sendSay(msg) ;
   } ;

   //======================================================================
   // low level command

   //--------------------------------------------------
   // turns

   Bool SimplePlayer::turnTo(const Flt adir, 
			     const Flt & cutoff) {
      /* <<< [Bun-0.3] <<< */

      //return turn(normalizeAngle(adir - mybdir())) ;

      /* === [Bun-0.3] === */

      return turn(normalizeAngle(adir - estSelf().bdir()),cutoff) ;

      /* >>> [Bun-0.3] >>> */
   } ;

   Bool SimplePlayer::turnToRel(const Vec & rpos, 
				const Flt & cutoff) {
      return turnTo(rpos.angle(),cutoff) ;
   } ;

   Bool SimplePlayer::turnToAbs(const Vec & apos, 
				const Flt & cutoff) {

      /* <<< [Bun-0.3] <<< */

      //return turnToRel(mypos().diffTo(apos)) ;

      /* === [Bun-0.3] === */

      return turnToRel(estSelfPos().diffTo(apos),cutoff) ;

      /* >>> [Bun-0.3] >>> */
   } ;

   /* <<< [Bun-0.4] <<< */

   Bool SimplePlayer::turnBackToAbs(const Vec & apos, 
				    const Flt & cutoff) {

      return turnToRel(-(estSelfPos().diffTo(apos)),cutoff) ;

   } ;

   /* >>> [Bun-0.4] >>> */

   //--------------------------------------------------
   // kicks

   Bool SimplePlayer::kickTo(const Flt & power, const Flt & adir) {

     WorldModel::Action action;
     WorldModel::State cstate = worldmodel.estState(stime().count);

     const FieldState::Object & cball = cstate.ball() ;
     const FieldState::Object & self  = cstate.self() ;

      Flt angdiff  = normalizeAngle((cball.pos() - self.pos()).angle() - self.bdir()) ;
      Flt distdiff = (cball.pos() - self.pos()).norm() - worldmodel.playerSize - worldmodel.ballSize;

      if(distdiff < worldmodel.kickableMargin * 1.05) {
	
	Flt tmppower = power / worldmodel.kickPowerFactor(distdiff,angdiff) ;
	Flt tmpdir   = normalizeAngle(adir - self.bdir()) ;

	return kick(tmppower, tmpdir) ;
     } 
   };

  /*
   Bool SimplePlayer::kickTo(const Flt & power, const Flt & adir) {
      return kick(power, normalizeAngle(adir - estSelf().bdir())) ;
   } ;
  */

   Bool SimplePlayer::kickToRel(const Flt & power, const Vec & rpos) {
      return kickTo(power, rpos.angle()) ;
   } ;

   Bool SimplePlayer::kickToAbs(const Flt & power, const Vec & apos) {
      /* <<< [Bun-0.3] <<< */

      //return kickToRel(power, mypos().diffTo(apos)) ;

      /* === [Bun-0.3] === */

      return kickToRel(power, estSelfPos().diffTo(apos)) ;

      /* >>> [Bun-0.3] >>> */
   } ;

   //--------------------------------------------------
   // gcatch

   Bool SimplePlayer::gcatchTo(const Flt & adir) {
      /* <<< [Bun-0.3] <<< */

      //return gcatch(normalizeAngle(adir - mybdir())) ;

      /* === [Bun-0.3] === */

      return gcatch(normalizeAngle(adir - estSelf().bdir())) ;

      /* >>> [Bun-0.3] >>> */
   } ;

   Bool SimplePlayer::gcatchToRel(const Vec & rpos) {
      return gcatchTo(rpos.angle()) ;
   } ;

   Bool SimplePlayer::gcatchToAbs(const Vec & apos) {
      /* <<< [Bun-0.3] <<< */

      //return gcatchToRel(mypos().diffTo(apos)) ;

      /* === [Bun-0.3] === */

      return gcatchToRel(estSelfPos().diffTo(apos)) ;

      /* >>> [Bun-0.3] >>> */
   } ;

   //--------------------------------------------------
   // neck

   Bool SimplePlayer::headTo(const Angle & adir) {
      /* <<< [Bun-0.3] <<< */

      //return head(normalizeAngle(adir - myhdir())) ;

      /* === [Bun-0.3] === */

      return head(normalizeAngle(adir - estSelf().hdir())) ;

      /* >>> [Bun-0.3] >>> */

   } ;

   Bool SimplePlayer::headToRel(const Vec & rpos) {
      return headTo(rpos.angle()) ;
   } ;

   Bool SimplePlayer::headToAbs(const Vec & apos) {
      /* <<< [Bun-0.3] <<< */

      //return headToRel(mypos().diffTo(apos)) ;

      /* === [Bun-0.3] === */

      return headToRel(estSelfPos().diffTo(apos)) ;

      /* >>> [Bun-0.3] >>> */

   } ;


   //--------------------------------------------------
   // trap & kieck
   Bool SimplePlayer::trapBall(Flt hardness) {

      WorldModel::Action action ;
      WorldModel::State cstate = worldmodel.estState(stime().count) ;
	       
      Vec bvel = cstate.self().pos() - cstate.ball().pos() ;
      bvel.normalize(hardness) ;

      Bool r = worldmodel.estimateBallBackActionByBallVel(action, bvel,cstate) ;
      if(r) {
	 return kick(action.kickPower(),action.kickDir()) ;
      } else {
	 return False ;
      }
   }
	 
   Bool SimplePlayer::trapKickTo(const Flt & power, const Flt & adir, Flt hardness) {
      trapBall(hardness) ;
      waitNextBestComTime() ;
      return kickTo(power,adir) ;
   }

   Bool SimplePlayer::trapKickToRel(const Flt & power, const Vec & rpos,
				    Flt hardness) {
      trapBall(hardness) ;
      waitNextBestComTime() ;
      return kickToRel(power,rpos) ;
   }

   Bool SimplePlayer::trapKickToAbs(const Flt & power, const Vec & apos,
				    Flt hardness) {
      trapBall(hardness) ;
      waitNextBestComTime() ;
      return kickToAbs(power,apos) ;
   }

   //--------------------------------------------------
   // swing ball 1

   Bool SimplePlayer::swingBall1(Flt radius, Flt dirstep) {
      WorldModel::State cstate 
	 = worldmodel.estState(stime().count) ;

      Vec dpos = cstate.ball().pos() - cstate.self().pos() ;
      Flt ndir = dpos.angle() + dirstep ;
      Vec npos = cstate.self().pos() + Geo2D::polar2Vec(radius,ndir) ;

      WorldModel::Action action ;
      Bool r = worldmodel.estimateBallBackActionByBallPos(action, npos,cstate) ;
      if(r) return kick(action.kickPower(),action.kickDir()) ;
      else return False ;
   } ;

   //--------------------------------------------------
   // swing ball N

   Bool SimplePlayer::swingBallN(Int n, Flt radius, Flt dirstep, 
				 Bool stopp, Int waitmargin) {
      for(Int i = 0 ; i < n ; i++) {
	 if(!kickable()) return False ;
	 waitNextBestComTime(waitmargin) ;
	 if(!swingBall1(radius,dirstep)) return False ;
      }
      
      if(stopp) 
	 return trapBall(0.0) ;
      else
	 return True ;
   } ;

   //--------------------------------------------------
   // swing ball to a target dir

       /* if shorterAngle == True, and dirstep is positive, then
	* use the same direction of shorter angle.  
	* if dirstep is negative, then use another direction of shorter angle.
	* if shorterAngle == False and dirstep is positive, then clockwise
	* else counter-clockwise
	*/

   Bool SimplePlayer::swingBallTo(Flt tdir, Flt radius, Flt dirstep, 
				  Bool shorterAngle,
				  Bool stopp, Int waitmargin) {

      while(True) {
	 if(!kickable()) return False ;

	 waitNextBestComTime(waitmargin) ;
	 WorldModel::State cstate 
	    = worldmodel.estState(stime().count) ;
	 Vec dpos = cstate.ball().pos() - cstate.self().pos() ;
	 Flt bdir = dpos.angle() ;

	 Flt ddir = normalizeAngle(tdir - bdir) ;

	 if(Abs(ddir) < 5.0)  break ; // !!!!! cutoff threashold
	 
	 if(Abs(ddir) < Abs(dirstep)) {
	    swingBall1(radius,ddir) ;
	 } else {
	    if(shorterAngle) {
	       swingBall1(radius,Sign(ddir) * dirstep) ;
	    } else {
	       swingBall1(radius,dirstep) ;
	    }
	 }
      }

      if(stopp) 
	 return trapBall(0.0) ;
      else
	 return True ;
   } ;

   //--------------------------------------------------
   // swing ball to a target relative pos

   Bool SimplePlayer::swingBallToRel(Pos rpos, Flt radius, Flt dirstep, 
				     Bool shorterAngle,
				     Bool stopp, Int waitmargin) {
      return swingBallTo(rpos.angle(),radius,dirstep,
			 shorterAngle,stopp,waitmargin) ;
   } ;
   
   //--------------------------------------------------
   // swing ball to a target abs pos

   Bool SimplePlayer::swingBallToAbs(Pos apos, Flt radius, Flt dirstep, 
				     Bool shorterAngle,
				     Bool stopp, Int waitmargin) {
      return 
	 swingBallToRel(
	    worldmodel.estState(stime().count).self().pos().diffTo(apos),
	    radius,dirstep,shorterAngle,stopp,waitmargin) ;
   } ;
   

   //--------------------------------------------------
   // dribble

   Bool SimplePlayer::dribble1To(Flt tdir, Flt power, 
				 Flt keepdir, Flt keepdist, Flt dirstep,
				 Int waitmargin) {
#if 0      
      Flt dashpower = Min(100.0, power) ;
      Flt ratio = 
	 0.2 * 			// !!!!
	 worldmodel.dashPowerRate / worldmodel.kickPowerRate ;
      ITK_DBG(ratio) ;
      Flt kickpower = dashpower * ratio ; 
      
      if(!swingBallTo(keepdir, keepdist, dirstep, True, waitmargin))
	 return False ;

      waitNextBestComTime(waitmargin) ;
      if(!kickTo(kickpower,tdir)) return False ;

      waitNextBestComTime(waitmargin) ;
      turnTo(tdir) ;

      waitNextBestComTime(waitmargin) ;
      dash(dashpower) ;

      if(!swingBallTo(keepdir, keepdist, dirstep, True, waitmargin))
	 return False ;

      return True ;
#endif
      if(!swingBallTo(tdir,0.8,20,True,waitmargin)) return False ;
      
      waitNextBestComTime(waitmargin) ;
      turnTo(tdir) ;
      headTo(tdir) ;

      waitNextBestComTime(waitmargin) ;
      dash(power) ;

      if(!swingBallTo(tdir, 0.5, 20, True, waitmargin)) ;
	 return False ;

      return True ;

   } ;

} ;
   
 
   

