// -*- Mode:C++ -*-
//Header:
//File: WorldModel.cc
//Title: WorldModel.cc
//Author: NODA, Itsuki
//Date: 2001/07/27
//

//ModifyHistory:
// 2001/07/27: Start to create this file
//EndModifyHistory:

#include "WorldModel.h"

namespace Soccer {

   //======================================================================
   // WorldModel:: class State

   //======================================================================
   // class WorldModel

   //------------------------------------------------------------
   // WorldModel:: estState
  // worktime ᤹ɬפϤʤΤ
  // estState ϲäƤΤ?
  //  t λ FieldState ֤
   WorldModel::State & WorldModel::estState(Int t, Bool lockp) {
      State * r = ITK_NULLPTR ;

      if(lockp) mutex.lock() ;
      {
	 if(t < worktime - stateN) {
	    ITK_WRN(*me() << ": Too old world state") ;
	    ITK_DBG(t) ;
	    ITK_DBG(worktime) ;
	    ITK_DBG(stateN) ;
	 }

	 // checking old state
	 if      (t < worktime) {
	    r = &state(t) ;
	 }
	 else {
	    if (t > worktime) {
	       estState(t-1,False) ;
	    }
	    r = &estimate1(t,False) ;
	 }
      }
      if(lockp) mutex.unlock();
      
      return *r ;
   }
      
   //------------------------------------------------------------
   // WorldModel:: estimate1
   //   estimate state(t) from state(t-1)
   //   suppose state(t-1) is calculated
   //   return state(t)
  //%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  //%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  //%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
   WorldModel::State & WorldModel::estimate1(Int t, Bool lockp) {
      State * r ;

      if(lockp) mutex.lock() ;
      {
	 State & pstate = state(t-1) ;
	 State & cstate = state(t) ;
	 Action & paction = action(t-1) ;
	 r = &cstate ;

	 cstate.reset() ;

	 cstate.time() = t ;

	 estimateFore(cstate,pstate,paction) ;

	 cstate.updateTime().update() ;

	 worktime = t + 1 ;
      }
      if(lockp) mutex.unlock() ;
	    
      return *r ;
   } ;
  //%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  //%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


   //------------------------------------------------------------
   // WorldModel:: put sensored data

   void WorldModel::mergeSensedData(Int t, const FieldState & fs, Bool lockp) {
      if(lockp) mutex.lock() ;
      {

	 /* <<< [Bun-0.2] <<< */
	 
	 //mergeSensedDataSelf(t, fs, False) ;
	 //mergeSensedDataBall(t, fs, False) ;
	 
	 /* === [Bun-0.2] === */
	 
	 mergeSensedDataSelf2(t, fs, False) ;
	 mergeSensedDataBall2(t, fs, False) ;
	 
	 /* >>> [Bun-0.2] >>> */

	 state(t).updateTime().update() ;

	 worktime = t + 1 ;
      } 
      if(lockp) mutex.unlock() ;
   } ;

   //----------------------------------------
   // WorldModel:: put sensored data (self)

   void WorldModel::mergeSensedDataSelf(Int t, 
					const FieldState & fs, 
					Bool lockp) {
      if(lockp) mutex.lock() ;
      {
	 if(fs.self().posConf > 0.0) {
	    State & cstate = estState(t,False) ;
	    Object & self = cstate.self() ;

	    /* direction is copied from sensor value */
	    self.bdir() = fs.self().bdir() ;
	    self.hdir() = fs.self().hdir() ;

	    /* merge pos */
	    Bool useSensor = False ;

	    //<<<[Bun-0.1]
	    // body operations are moved to sub-functions

	    mergeSensedDataSelfPos(t,fs,cstate,self,useSensor) ;
	    mergeSensedDataSelfVel(t,fs,cstate,self,useSensor) ;

	    //>>>[Bun-0.1]
	 }

      } 
      if(lockp) mutex.unlock() ;
   } ;

   //<<<[Bun-0.1]
   //--------------------
   // WorldModel:: put sensored data (self)
   
   void WorldModel::mergeSensedDataSelfPos(Int t, const FieldState & fs,
					   State & cstate, Object & self,
					   Bool & useSensor) {
      if(cstate.selfAuthority() == State::Auth_Estimate) {
	 Pos posdiff = self.pos() - fs.self().pos() ;
	 if(posdiff.norm() > useSensorCutoff) {
	    useSensor = True ;
	 } else {
	    self.pos() = 
	       fs.self().pos() * mergeSensorRate +
	       self.pos() * (1.0 - mergeSensorRate) ;
	    self.posConf = 
	       fs.self().posConf * mergeSensorRate +
	       self.posConf * (1.0 - mergeSensorRate) ;
	    self.velConf = 
	       fs.self().velConf * mergeSensorRate +
	       self.velConf * (1.0 - mergeSensorRate) ;
		  
	    cstate.selfAuthority() = State::Auth_Mixed ;
	    cstate.selfNoSensor() = 0 ;
	    cstate.selfSensorMixRatio() = mergeSensorRate ;
	 }
      }

      if(useSensor) {
	 self.pos() = fs.self().pos() ;
	 self.posConf = fs.self().posConf ;
	 self.velConf = fs.self().velConf ;
	       
	 cstate.selfAuthority() = State::Auth_Sensor ;
	 cstate.selfNoSensor() = 0 ;
	 cstate.selfSensorMixRatio() = 1.0 ;
      }
   } ;

   //--------------------
   // WorldModel:: put sensored data (self)
   
   void WorldModel::mergeSensedDataSelfVel(Int t, const FieldState & fs,
					   State & cstate, Object & self,
					   Bool & useSensor) {
      self.vel() =
	 (self.pos() - estState(t-1,False).self().pos()) 
	 * decayPlayerVel ;

      if(cstate.myspeed() >= 0.0) {
	 cstate.mergeMyspeed(cstate.myspeed()) ;
      }

      if(self.vel().norm() > playerSpeedMax) {
	 self.vel().set(0,0) ;
	 self.velConf = 0.0 ;
      }
   } ;
   
   //>>>[Bun-0.1]

   //----------------------------------------
   // WorldModel:: put sensored data (ball)

   void WorldModel::mergeSensedDataBall(Int t, 
					const FieldState & fs, 
					Bool lockp) {
      if(lockp) mutex.lock() ;
      {
	 if(fs.ball().posConf > 0.0) {
	    State & cstate = estState(t,False) ;
	    Object & ball = cstate.ball() ;

	    Bool useSensor = False ;
	    Bool shortRange = False ;

	    //<<<[Bun-0.1] 
	    // body operations are moved to sub-functions

	    mergeSensedDataBallPos(t,fs,cstate,ball,useSensor,shortRange) ;

	    mergeSensedDataBallVel(t,fs,cstate,ball,useSensor,shortRange) ;

	    //>>>[Bun-0.1]
	 }

      } 
      if(lockp) mutex.unlock() ;
   } ;

   //<<<[Bun-0.1]
   //--------------------
   // WorldModel:: put sensored data (ball)

   void WorldModel::mergeSensedDataBallPos(Int t, const FieldState & fs,
					   State & cstate,
					   Object & ball,
					   Bool & useSensor, 
					   Bool & shortRange) {
      Flt bdist = (fs.ball().pos() - fs.self().pos()).norm() ;
      if(bdist < shortRangeThreshold) shortRange = True ;

      Flt sensorPosConf = 1.0 / fs.ball().sdist() ;

      if(cstate.ballAuthority() == State::Auth_Estimate) {
	 Pos posdiff = ball.pos() - fs.ball().pos() ;
	 Flt cutoff = 
	    useSensorCutoff + fs.ball().sdist() * distQuantizeStep ;

	 if(posdiff.norm() > cutoff) { /* if estimated ball pos is
					* too different from the sensor
					* data, then use only sensor
					* data. */
	    useSensor = True ;
	 } else {
	    Flt w0 = ball.posConf + sensorPosConf ;
	    Flt we = ball.posConf/w0 ; // instead of "mergeSensorRate" 
	    Flt ws = sensorPosConf/w0 ;  // instead of "mergeSensorRate" 
	    ball.pos() =
	       fs.ball().pos() * ws + ball.pos() * we ;
	    ball.posConf =
	       fs.ball().posConf * ws + ball.posConf * we ;
	    ball.velConf =
	       fs.ball().velConf * ws + ball.velConf * we ;

	    cstate.ballAuthority() = State::Auth_Mixed ;
	    cstate.ballNoSensor() = 0 ;
	    cstate.ballSensorMixRatio() = ws ;
	 }

	 if(useSensor || shortRange) { /* in the case of useSensor flag,
					* or shortRange, discard estimate
					* value and use sensor data */
					    
	    ball.pos() = fs.ball().pos() ;
	    //ball.posConf = fs.ball().posConf ;   /* use sdist() value */
	    ball.posConf = sensorPosConf ;
	    ball.velConf = fs.ball().velConf ;

	    cstate.ballAuthority() = State::Auth_Sensor ;
	    cstate.ballNoSensor() = 0 ;
	    cstate.ballSensorMixRatio() = 1.0 ;
	 }
	 
      }
   } ;

   //--------------------
   // WorldModel:: put sensored data (ball)

   void WorldModel::mergeSensedDataBallVel(Int t, const FieldState & fs,
					   State & cstate,
					   Object & ball,
					   Bool & useSensor, 
					   Bool & shortRange) {
      
      if(shortRange) {		/* in the case of short range
				 * use relative position to estimate
				 * velocity */
	 Pos pbposdiff = 
	    estState(t-1,False).ball().pos()
	    - estState(t-1,False).self().pos() ;
	 Pos cbposdiff =
	    cstate.ball().pos()
	    - cstate.self().pos() ;
	 ball.vel() = 
	    (cbposdiff - pbposdiff) * decayBallVel 
	    + cstate.self().vel() ;
      } 
      else {			/* otherwise, use weighted average
				 * of past N cycles */
	 Vec v ; v.set(0,0) ;
	 Flt n = 0.0 ;
	 Flt decay = 1.0 ;
	 Flt conf = 1.0 ;
	 //<<<[Bun-0.1]
	 //Flt confDecay = 1.0 ; // !!!
	 Flt confDecay = 0.8 ; 
	 //UInt N = 3 ;
	 UInt N = 5 ;
	 //UInt N = 5 ;
	 //>>>[Bun-0.1]
	 for(UInt i = 0 ; i < N ; i++) {
	    Vec dv = (estState(t-i,False).ball().pos() - 
		      estState(t-i-1,False).ball().pos()) * decay * conf;
	    v += dv ;
	    n += conf ;

	    conf *= confDecay ; 
	    decay *= decayBallVel ;
	 }

	 ball.vel() = v * (decayBallVel / n) ;
      }

      if(ball.vel().norm() > ballSpeedMax) {
	 ball.vel().set(0,0) ;
	 ball.velConf = 0.0 ;
      }
   } ;
   
   //>>>[Bun-0.1]

   /* <<< [Bun-0.2] <<< */

   //----------------------------------------
   // WorldModel:: put sensored data (self) (Kalman Filter version)

   void WorldModel::mergeSensedDataSelf2(Int t, 
					 const FieldState & fs, 
					 Bool lockp) {
      if(lockp) mutex.lock() ;
      {
	 if(fs.self().posConf > 0.0) {
	    State & cstate = estState(t,False) ;
	    Object & self = cstate.self() ;

	    /* direction is copied from sensor value */
	    self.bdir() = fs.self().bdir() ;
	    self.hdir() = fs.self().hdir() ;

	    /* Kalman Filter */

	    ValVector &Xf = selfKFilter().vXf() ;
	    Pos acc(0.0,0.0) ;
	    
	    Xf.val(0) = self.pos().x ; Xf.val(1) = self.pos().y ;
	    Xf.val(2) = self.vel().x ; Xf.val(3) = self.vel().y ;
	    Xf.val(4) = acc.x        ; Xf.val(5) = acc.y        ;

	    selfKFilter().wPf() = cstate.selfKFilterPf() ;

	    /* <<< [Bun-0.3] <<< */
	    /* ʬ򥻥å 
	     */

	    selfKFilter().setParamObserv(fs.myposVar()) ;

	    /* >>> [Bun-0.3] >>> */

	    selfKFilter().mergeSensor(fs.self().pos().x,
				      fs.self().pos().y) ;
	    
	    cstate.selfKFilterPf() = selfKFilter().wPf() ;
	    
	    self.pos() = Pos(Xf.val(0),Xf.val(1)) ;
	    self.vel() = Pos(Xf.val(2),Xf.val(3)) ;

	    /* end Kalman Filter */
	    
	    if(cstate.myspeed() >= 0.0) {
	       cstate.mergeMyspeed(cstate.myspeed()) ;
	    }

	    cstate.selfAuthority() = State::Auth_Mixed ;
	    cstate.selfNoSensor() = 0 ;
	    cstate.selfSensorMixRatio() = 0.5 ; // ???

	    if(self.vel().norm() > playerSpeedMax) {
	       self.vel().set(0,0) ;
	       self.velConf = 0.0 ;
	    }
	 }

      } 
      if(lockp) mutex.unlock() ;
   } ;


   //----------------------------------------
   // WorldModel:: put sensored data (ball) (Kalman Filter version)

   void WorldModel::mergeSensedDataBall2(Int t, 
					 const FieldState & fs, 
					 Bool lockp) {
      if(lockp) mutex.lock() ;
      {
	 if(fs.ball().posConf > 0.0) {
	    State & cstate = estState(t,False) ;
	    Object & ball = cstate.ball() ;

	    /* Kalman Filter */

	    ValVector & Xf = ballKFilter().vXf() ;
	    Pos acc(0.0,0.0) ;

	    Xf.val(0) = ball.pos().x ; Xf.val(1) = ball.pos().y ;
	    Xf.val(2) = ball.vel().x ; Xf.val(3) = ball.vel().y ;
	    Xf.val(4) = acc.x        ; Xf.val(5) = acc.y        ;

	    ballKFilter().wPf() = cstate.ballKFilterPf() ;

	    /* <<< [Bun-0.3] <<< */
	    /* ʬ򥻥å 
	     */

	    ballKFilter().setParamObserv(fs.ballposVar()) ;

	    /* >>> [Bun-0.3] >>> */

	    ballKFilter().mergeSensor(fs.ball().pos().x,
				      fs.ball().pos().y) ;
	    
	    cstate.ballKFilterPf() = ballKFilter().wPf() ;
	    
	    ball.pos() = Pos(Xf.val(0),Xf.val(1)) ;
	    ball.vel() = Pos(Xf.val(2),Xf.val(3)) ;
	    
	    cstate.ballAuthority() = State::Auth_Mixed ;
	    cstate.ballNoSensor() = 0 ;
	    cstate.ballSensorMixRatio() = 0.5 ; // ???

	 }

      } 
      if(lockp) mutex.unlock() ;
   } ;
   /* >>> [Bun-0.2] >>> */


   //----------------------------------------
   // WorldModel::State merge speed data

   void WorldModel::State::mergeMyspeed(Flt speed) {
      myspeed() = speed ;
      self().vel().normalize(myspeed()) ;
   }

//--------------------
// old one [2001.09.20]
#if 0
   void WorldModel::mergeSensedData(Int t, const FieldState & fs, Bool lockp) {
      if(lockp) mutex.lock() ;

      {
	 State & pstate = estState(t-1,False) ;
	 State & cstate = estState(t,False) ;

	 if(fs.self().posConf > 0.0) {
	    cstate.self().bdir() = fs.self().bdir() ;
	    cstate.self().hdir() = fs.self().hdir() ;
	 }
	 if(fs.self().posConf > cstate.self().posConf) {
	    cstate.self().pos() = fs.self().pos() ;
	    cstate.self().posConf = fs.self().posConf ;
	 }
	 if(fs.self().velConf > cstate.self().velConf) {
	    cstate.self().vel() = fs.self().vel() ;
	    cstate.self().velConf = fs.self().velConf ;
	 }

	 if(fs.ball().posConf > cstate.ball().posConf) {
	    cstate.ball().pos() = fs.ball().pos() ;
	    cstate.ball().posConf = fs.ball().posConf ;

	    Vec v ; v.set(0,0) ;
	    Flt n = 0.0 ;
	    Flt decay = 1.0 ;
	    Flt conf = 1.0 ;
	    Flt confDecay = 1.0 ; // !!!
	    for(UInt i = 0 ; i < 3 ; i++) {
	       v += (estState(t-i,False).ball().pos() - 
		     estState(t-i-1,False).ball().pos()) * decay * conf ;
	       n += conf ;

	       conf *= confDecay ; 
	       decay *= decayBallVel ;
	    }

	    cstate.ball().vel() = v * (decayBallVel / n) ;

	    cstate.ball().velConf = 
	       cstate.ball().posConf * pstate.ball().posConf ;
	 } ;

	 worktime = t + 1 ;

      } 
      if(lockp) mutex.unlock() ;
   } ;
#endif

   //------------------------------------------------------------
   // WorldModel:: estimate Forward
   //     state should be reseted and set time().
   
   void WorldModel::estimateFore(State & cstate, 
				 const State & pstate, 
				 const Action & paction) {
      /* <<< [Bun-0.2] <<< */

      //estimateSelfFore(cstate,pstate,paction) ;
      //estimateBallFore(cstate,pstate,paction) ;

      /* === [Bun-0.2] === */

      estimateSelfFore2(cstate,pstate,paction) ;      
      estimateBallFore2(cstate,pstate,paction) ;

      /* >>> [Bun-0.2] >>> */
   } ;

   //----------------------------------------
   // WorldModel:: estimate Self Forward

   void WorldModel::estimateSelfFore(State & cstate, 
				     const State & pstate, 
				     const Action & paction) {

      Object & self = cstate.self() ;

      self.copyFrom(pstate.self()) ;

      if(paction.varidp()) {
	 if      (paction.dashp()) {
	    Flt effort = 1.0 ;
	    Flt realpow = paction.dashPower() * effort * dashPowerRate ;
	    Pos acc ; acc.setByPolar(realpow,self.bdir()) ;
	    self.vel() += acc ;
	 } 
	 else if (paction.turnp()) {
	    Flt realpow = (paction.turnPower() 
			   / (1.0 + inertia * self.vel().norm())) ;
	    self.bdir() += realpow ; letNormalizeAngle(self.bdir()) ;
	    self.hdir() += realpow ; letNormalizeAngle(self.hdir()) ;
	 }
	 if(paction.headp()) {
	    self.hdir() += paction.headPower() ;
	    Flt dangle = normalizeAngle(self.hdir() - self.bdir()) ;
	    if(dangle >  neckMax) 
	       self.hdir() = normalizeAngle(self.bdir() + neckMax) ;
	    if(dangle < -neckMax)
	       self.hdir() = normalizeAngle(self.bdir() - neckMax) ;
	 }
      }

      self.pos() += self.vel() ;
      self.vel() *= decayPlayerVel ;      
      self.posConf *= decayConf ;
      self.velConf *= decayConf ;

      cstate.selfAuthority() = State::Auth_Estimate ;
      cstate.selfNoSensor() = pstate.selfNoSensor() + 1;
      cstate.selfSensorMixRatio() = pstate.selfSensorMixRatio() * 0.5 ; 
   } ;
	    
   //----------------------------------------
   // WorldModel:: estimate Ball Forward
   void WorldModel::estimateBallFore(State & cstate, 
				     const State & pstate, 
				     const Action & paction) {

      Object & ball = cstate.ball() ;
      Object & self = cstate.self() ;

      ball.copyFrom(pstate.ball()) ;

      if(paction.varidp()) {
	 if (paction.kickp()) {
	    Flt angdiff  = normalizeAngle((ball.pos() - self.pos()).angle() - self.bdir()) ;
	    Flt distdiff = (ball.pos() - self.pos()).norm() - playerSize - ballSize ;

	    if(distdiff < kickableMargin) {
	       Flt realpower = 
		  paction.kickPower() * kickPowerFactor(distdiff,angdiff) ;

	       Pos acc ; 
	       acc.setByPolar(realpower,paction.kickDir() + self.bdir()) ;

	       ball.vel() += acc ;
	    }
	 }
      }
	    
      ball.pos() += ball.vel() ;
      ball.vel() *= decayBallVel ;
      ball.posConf *= decayConf ;
      ball.velConf *= decayConf ;
      
      cstate.ballAuthority() = State::Auth_Estimate ;
      cstate.ballNoSensor() = pstate.ballNoSensor() + 1;
      cstate.ballSensorMixRatio() = pstate.ballSensorMixRatio() * 0.5 ; 
   } ;

   /* <<< [Bun-0.2] <<< */

   //----------------------------------------
   // WorldModel:: estimate Self Forward (Kalman Filter version)
  /*
    
   */
   void WorldModel::estimateSelfFore2(State & cstate, 
				      const State & pstate, 
				      const Action & paction) {

      Object & self = cstate.self() ;

      self.copyFrom(pstate.self()) ;

      Pos acc(0.0,0.0) ;

      /* encount command effect */

      if(paction.varidp()) {
	 if      (paction.dashp()) {
	    Flt effort = 1.0 ;
	    Flt realpow = paction.dashPower() * effort * dashPowerRate ;
	    acc.setByPolar(realpow,self.bdir()) ;
	 } 
	 else if (paction.turnp()) {
	    Flt realpow = (paction.turnPower() 
			   / (1.0 + inertia * self.vel().norm())) ;
	    self.bdir() += realpow ; letNormalizeAngle(self.bdir()) ;
	    self.hdir() += realpow ; letNormalizeAngle(self.hdir()) ;
	 }
	 if(paction.headp()) {
	    self.hdir() += paction.headPower() ;
	    Flt dangle = normalizeAngle(self.hdir() - self.bdir()) ;
	    if(dangle >  neckMax) 
	       self.hdir() = normalizeAngle(self.bdir() + neckMax) ;
	    if(dangle < -neckMax)
	       self.hdir() = normalizeAngle(self.bdir() - neckMax) ;
	 }
      }

      /* Kalman Filter */

      ValVector & Xf = selfKFilter().vXf() ;

      Xf.val(0) = self.pos().x ; Xf.val(1) = self.pos().y ;
      Xf.val(2) = self.vel().x ; Xf.val(3) = self.vel().y ;
      Xf.val(4) = acc.x        ; Xf.val(5) = acc.y        ;

      /* <<< [Bun-0.3] <<< */
      /* ®٤˱Υѥ᡼򥻥åȡ
       * noise  d = vel().norm() * rand Ȥơ
       * [-d,d] ΰʬۤʤΤǡʬ d^2 / 3.0
       */
      selfKFilter().setParamPredNoise(
	 0.0, 				   	  /* sigma pos */
	 sq(self.vel().norm() * playerRand)/3.0, /* sigma vel. */
	 playerSigmaAcc) ;		   	  /* acc. */
      /* >>> [Bun-0.3] >>> */

      selfKFilter().wPf() = pstate.selfKFilterPf() ;

      selfKFilter().estCycle() ;

      cstate.selfKFilterPf() = selfKFilter().wPf() ;
	    
      self.pos() = Pos(Xf.val(0), Xf.val(1)) ;
      self.vel() = Pos(Xf.val(2), Xf.val(3)) ;

      /* end Kalman Filter */

      self.posConf *= decayConf ;
      self.velConf *= decayConf ;

      cstate.selfAuthority() = State::Auth_Estimate ;
      cstate.selfNoSensor() = pstate.selfNoSensor() + 1;
      cstate.selfSensorMixRatio() = pstate.selfSensorMixRatio() * 0.5 ; 
   } ;
	    
   //----------------------------------------
   // WorldModel:: estimate Ball Forward (Kalman Filter version)

   void WorldModel::estimateBallFore2(State & cstate, 
				     const State & pstate, 
				     const Action & paction) {

      Object & ball = cstate.ball() ;
      //Object & self = cstate.self() ;

      ball.copyFrom(pstate.ball()) ;

      /* Kalman Filter */

      ValVector & Xf = ballKFilter().vXf() ;

      Pos acc(0.0,0.0) ;	// !!! may be changable ???

      Xf.val(0) = ball.pos().x ; Xf.val(1) = ball.pos().y ;
      Xf.val(2) = ball.vel().x ; Xf.val(3) = ball.vel().y ;
      Xf.val(4) = acc.x        ; Xf.val(5) = acc.y        ;

      /* <<< [Bun-0.3] <<< */
      /* ®٤˱Υѥ᡼򥻥åȡ
       * noise  d = vel().norm() * rand Ȥơ
       * [-d,d] ΰʬۤʤΤǡʬ d^2 / 3.0
       */
      ballKFilter().setParamPredNoise(
	 0.0, 				   	/* sigma pos */
	 sq(ball.vel().norm() * ballRand)/3.0, /* sigma vel. */
	 ballSigmaAcc) ;			/* acc. */
      /* >>> [Bun-0.3] >>> */

      ballKFilter().wPf() = pstate.ballKFilterPf() ;

      ballKFilter().estCycle() ;

      cstate.ballKFilterPf() = ballKFilter().wPf() ;
	    
      ball.pos() = Pos(Xf.val(0), Xf.val(1)) ;
      ball.vel() = Pos(Xf.val(2), Xf.val(3)) ;

      /* end Kalman Filter */

      ball.posConf *= decayConf ;
      ball.velConf *= decayConf ;

      cstate.ballAuthority() = State::Auth_Estimate ;
      cstate.ballNoSensor() = pstate.ballNoSensor() + 1;
      cstate.ballSensorMixRatio() = pstate.ballSensorMixRatio() * 0.5 ; 
   } ;

   /* >>> [Bun-0.2] >>> */

   //----------------------------------------
   // WorldModel:: estimate Ball backword to realize ball velocity

   Bool WorldModel::estimateBallBackActionByBallVel(Action &caction,
						    State & nstate,
						    const State & cstate) {
      Object & nball = nstate.ball() ;
      Vec nbvel = nball.vel() ;

      return estimateBallBackActionByBallVel(caction,nbvel,cstate) ;
   } ;

   Bool WorldModel::estimateBallBackActionByBallVel(Action &caction,
						    const Vec & nbvel,
						    const State & cstate) {
      const Object & cball = cstate.ball() ;
      const Object & self = cstate.self() ;

      const Vec cbvel = cball.vel() ;

      Flt angdiff  = normalizeAngle((cball.pos() - self.pos()).angle()
				    - self.bdir()) ;
      Flt distdiff = (cball.pos() - self.pos()).norm() - playerSize - ballSize;

      // +0.05 餤;͵ OK?
      if(distdiff < kickableMargin * 1.05) {

	 Vec acc = nbvel - cbvel ;
	 Flt realpower = acc.norm() ;
	 Flt absdir = acc.angle() ;

	 Flt power = realpower / kickPowerFactor(distdiff,angdiff) ;
	 Flt dir = normalizeAngle(absdir - self.bdir()) ;

	 caction.varidp() = True ;
	 caction.setKick(power, dir);

	 return True ;
      } 
      else {
	 caction.varidp() = False ;
	 caction.kickp() = False ;
	 caction.kickPower() = 0 ;
	 caction.kickDir() = 0 ;

	 return False ;
      }
   }

   //----------------------------------------
   // WorldModel:: estimate Ball backword to realize ball position

   Bool WorldModel::estimateBallBackActionByBallPos(Action &caction,
						    State & nstate,
						    const State & cstate) {
      Object & nball = nstate.ball() ;
      Vec nbpos = nball.pos() ;

      return estimateBallBackActionByBallPos(caction,nbpos,cstate) ;
   } ;

   Bool WorldModel::estimateBallBackActionByBallPos(Action &caction,
						    const Vec & nbpos,
						    const State & cstate) {
      const Object & cball = cstate.ball() ;
      const Vec cbpos = cball.pos() ;

      Vec nbvel = nbpos - cbpos ;

      return estimateBallBackActionByBallVel(caction,nbvel,cstate) ;
   } ;

} ;


