/*
 * Header    :
 * File      : ChagammaBase.hxx 
 * Auther    : Kosuke Shinoda
 * Since     : "2002-03-11 22:27:48 kshinoda"
 * Time-stamp: "2002-03-15 14:10:52 kshinoda"
 * Comment   :
 * End       :
 */
/*
 * Copyright (C) 2001,2002 SHINODA, Kosuke. Jaist,Japan & CARC, AIST, JAPAN
 * Copyright (C) 2001 NODA, Itsuki, CARC, AIST, JAPAN
 */

#ifndef __Chagama_ChagamaBase_hxx__
#define __Chagama_ChagamaBase_hxx__

#include "SimplePlayer.h"
#include "Param.hxx"
#include "PlayPhase.hxx"

using namespace Itk;
using namespace Soccer;

//======================================================================
// class Chagamma


//======================================================================
// Option 

static VbsLevel VbsPlayPhase = 5;

class Chagamma0 : public SimplePlayer {

#define DFLT_TEAMNAME "chagamma"

  char * TeamName;
  char * ParamFile;
  
  char * SHost;
  Int SPort;
  Int OffCoachPort;
  Int CoachPort;

  Bool GoalieP;

  Int DebugLevel;

  Flt PlayerSpeedMax;

  Bool TestP;
  
  Flt Test2DashPower;

public:
  void initialize(){
    TeamName = DFLT_TEAMNAME ;
    ParamFile = DFLT_TEAMNAME ".param" ;
    SHost = "localhost" ;
    SPort        = 6000 ;
    OffCoachPort = 6001 ;
    CoachPort    = 6002 ;

    GoalieP     = False ;

    DebugLevel = 0 ;

    PlayerSpeedMax = 1.0 ;

    TestP = False ;
  
    Test2DashPower = 100 ;

    VbsPlayPhase = 5 ;
  };

  void getopt(int & argc, char ** & argv) {
    OptArg opt[] = {
      {"shost",     OptArg::cstringVal, &SHost, 0,   "sserver_hostname"},
      {"sport",	    OptArg::IntVal,     &SPort, 0,   "sserver_portnum"},
      {"goalie",    OptArg::OnOff,	&GoalieP,0,  "[on/off]"},
      {"team",	    OptArg::cstringVal, &TeamName,0, "teamname"},
      {"param",	    OptArg::cstringVal, &ParamFile,0, "parameter_filename"},
      {"debug",	    OptArg::IntVal,	&DebugLevel,0,  "debug_level [0-100]"},
      {ITK_NULLPTR}} ;
    myGetOpt(opt,argc,argv) ;
    ITK_SET_VBS(DebugLevel) ;
  }


      //============================================================
      // Chagamma0:: class TargetInfo
   public:
      class TargetInfo : public WithDescriber {
	 public:
	    Bool dribblep ;
	    Pos pos ;
	    Flt score ;
	    Flt power ;
	    char * name ;
	    Int aux ;

	 public:
	    virtual void describe(ostream & ostr, 
				  const Bool detailp = True) const {
	       ostr << "#TargetInfo[" << pos << (dribblep ? "(D)" : "(P)")
		    << ":score=" << score 
		    << "|" << name << "(" << aux << ")]" ;

	       if(detailp) ostr << endl ;
	    } ;
      } ;

      //============================================================
      // Chagamma0:: class CutPosInfo
   public:
      class CutPosInfo : public WithDescriber {
	 public:
	    Vec pos ;
	    Int maxdash ;

	 public:
	    CutPosInfo() {} ;
	    CutPosInfo(const Vec & p, const Int & m) {
	       pos = p ; maxdash = m ;
	    } ;

	 public:
	    virtual void describe(ostream & ostr, 
				  const Bool detailp = True) const {
	       ostr << "#CutPosInfo[" << pos << ",maxdash=" << maxdash << "]" ;
	    }
      } ;

      //------------------------------------------------------------
      // Chagamma0:: member
   public:
      Param param ;
      Param::RoleInfo * role ;

      PlayPhaseStack ppstack ;

      //------------------------------------------------------------
      // Chagamma0:: hear utility
   public:
      SimpleSexp::Heap hearItemHeap ;
      Mutex mtxHear ;

      //------------------------------------------------------------
      // Chagamma0:: say process and utility
   public:
      SimpleSexp::Heap sayItemHeap ;
      SimpleSexp * saylist ;
      Mutex mtxSay ;

      Thread * thrdSay ;

   public:
      void clearSay(Bool lockp = False) {
	 if(lockp) mtxSay.lock() ;
	 {
	    sayItemHeap.clear() ;
	    saylist = SimpleSexp::Nil ;
	 }
	 if(lockp) mtxSay.unlock() ;
      } ;

      SimpleSexp * addSayItem(SimpleSexp * s) {
	 return sayItemHeap.addFirst(s,saylist) ;
      } ;

      Thread * forkSay() ;
      static void * runSay(void *p) ;
      void * sayMainLoop() ;

#ifdef USE_NNN
      //------------------------------------------------------------
      // Chagamma0:: member: NnPredictor
   public:
      BallPredictor ballPredictor ;
#endif

      //------------------------------------------------------------
      // Chagamma0:: constructor
   public:
      Chagamma0() { 
#ifdef USE_NNN
	 ballPredictor.setup(worldmodel) ; 
#endif
	 initialize();
      } ;

      //------------------------------------------------------------
      // Chagamma0:: run 
   public:
      void run() ;
      void mainloop() ;

      void playDango() ;
      void playFixedPosition() ;
      
      void playGoalie() ;

      //------------------------------------------------------------
      // Chagamma0:: test run
   public:
      void playTest1() ;
      void playTest2() ;
      void playTest3() ;
      void playTest4() ;
      void playTest5() ;
      void playTest6() ;
      void playTest7() ;

      void playTest8() ;
      void playTest8_1() ;
      void playTest8_2() ;

      //------------------------------------------------------------
      // Chagamma0:: utility
   public:
      Bool checkKickOff() ;

      Bool checkPos() ;
      Bool checkBall() ;

      Flt badnessOfTarget(const Pos & target) ;
      Flt goodnessOfTarget(const Pos & target) ;
      Flt scoreOfTarget(const Pos & target) ;

      Bool compareTargetScore(TargetInfo & target, const Pos & newpos,
			      char * name = ITK_NULLPTR, Int aux = 0,
			      Bool dribblep = False);
      TargetInfo findBestTargetPos() ; // used in playFixedPosition

      Bool shouldChase() ;      // used in playFixedPosition

      Bool isClosestToBall(Int predict = 0, Bool ignoreGoalie = True) ;

      Pos homepos() ;

      //------------------------------------------------------------
      // Chagamma0:: goalie's utility
   public:
      Bool maybeSuccessCatch() {
	 if(isLeft(side)) {
	    return (playmode()->id == SC_WID_GoalieCatchL ||
		    playmode()->id == SC_WID_FreeKickL) ;
	 } else {
	    return (playmode()->id == SC_WID_GoalieCatchR ||
		    playmode()->id == SC_WID_FreeKickR) ;
	 }
      } ;
      
      //------------------------------------------------------------
      // Chagamma0:: find ball cut pos
   public:
      CutPosInfo cutBallPos() ;

      //------------------------------------------------------------
      // Chagamma0:: exactRun : do exact run to the dest position
   public:
      void exactRun(Pos destpos, Flt maxpower,
		    Bool needWaitFirstp = False, 
		    Int maxdash = 0) ;

      //------------------------------------------------------------
      // Chagamma0:: lookAroundWell
   public:
      void lookAroundWell() ;

      //------------------------------------------------------------
      // Chagamma0:: PlayPhase manipuration

      //----------------------------------------
      // Chagamma0:: push base
   public:
      PlayPhase & pushTopPlayPhase(Bool lockp = True) {
	 if(lockp) ppstack.lock() ;
	 PlayPhase & pp = ppstack.pushTop() ; 
	 if(lockp) {
	    ppstack.broadcast() ;
	    ppstack.unlock() ;
	 }
	 return pp ;
      } ;

      //----------------------------------------
      // Chagamma0:: push content
   public:
      PlayPhase & pushPlayPhase(const PlayPhase & p, Bool lockp = True) {
	 if(lockp) ppstack.lock() ;
	 PlayPhase & pp = ppstack.pushTop() ; 
	 pp = p ;
	 ITK_VBS(VbsPlayPhase,*this << "push PlayPhase:" << pp) ;
	 if(lockp) {
	    ppstack.broadcast() ;
	    ppstack.unlock() ;
	 }
	 return pp ;
      } ;
      
      //----------------------------------------
      // Chagamma0:: pop

      PlayPhase & popPlayPhase(Bool lockp = True) {
	 if(lockp) ppstack.lock() ;
	 PlayPhase & pp = ppstack.pop() ;
	 ITK_VBS(VbsPlayPhase,*this << "pop PlayPhase:" << pp) ;
	 if(lockp) {
	    ppstack.broadcast() ;
	    ppstack.unlock() ;
	 }
	 return pp ;
      } ;

      //----------------------------------------
      // Chagamma0:: push with Flt args
   public:
      PlayPhase & pushPlayPhase(PlayPhase::Type type, 
				Flt arg1,
				Flt arg2 = 0.0,
				Flt arg3 = 0.0,
				Bool lockp = True) {
	 PlayPhase pp ;
	 pp.type = type ;
	 pp.farg1() = arg1 ;
	 pp.farg2() = arg2 ;
	 pp.farg3() = arg3 ;

	 PlayPhase & r = pushPlayPhase(pp,lockp) ;

	 return r ;
      } ;

      //----------------------------------------
      // Chagamma0:: push with Int args
   public:
      PlayPhase & pushPlayPhase(PlayPhase::Type type, 
				Int arg1 = 0,
				Int arg2 = 0,
				Int arg3 = 0,
				Bool lockp = True) {
	 PlayPhase pp ;
	 pp.type = type ;
	 pp.iarg1() = arg1 ;
	 pp.iarg2() = arg2 ;
	 pp.iarg3() = arg3 ;

	 PlayPhase & r = pushPlayPhase(pp,lockp) ;

	 return r ;
      } ;

      //----------------------------------------
      // Chagamma0:: get current PlayPhase
   public:
      PlayPhase & currentPlayPhase(Bool lockp = True) {
	 if(lockp) ppstack.lock() ;
	 PlayPhase & pp = ppstack.top() ;
	 if(lockp) ppstack.unlock() ;
	 return pp ;
      } ;

      //----------------------------------------
      // Chagamma0:: change current PlayPhase
   public:
      PlayPhase & changePlayPhase(const PlayPhase & newpp, Bool lockp = True) {
	 if(lockp) ppstack.lock() ;
	 PlayPhase & pp = currentPlayPhase(False) ;
	 ITK_VBS(VbsPlayPhase,*this << "change PlayPhase from" << pp 
		 << " to " << newpp) ;
	 pp = newpp ;
	 ppstack.broadcast() ;
	 if(lockp) ppstack.unlock() ;
	 return pp ;
      } ;

      //----------------------------------------
      // Chagamma0:: change current PlayPhase with Flt args
   public:
      PlayPhase & changePlayPhase(PlayPhase::Type type,
				  Flt arg1,
				  Flt arg2 = 0.0,
				  Flt arg3 = 0.0,
				  Bool lockp = True) {
	 PlayPhase pp ;
	 pp.type = type ;
	 pp.farg1() = arg1 ;
	 pp.farg2() = arg2 ;
	 pp.farg3() = arg3 ;

	 PlayPhase & r = changePlayPhase(pp,lockp) ;

	 return r ;
      } ;

      //----------------------------------------
      // Chagamma0:: change current PlayPhase with Int args
   public:
      PlayPhase & changePlayPhase(PlayPhase::Type type,
				  Int arg1 = 0,
				  Int arg2 = 0,
				  Int arg3 = 0,
				  Bool lockp = True) {
	 PlayPhase pp ;
	 pp.type = type ;
	 pp.iarg1() = arg1 ;
	 pp.iarg2() = arg2 ;
	 pp.iarg3() = arg3 ;

	 PlayPhase & r = changePlayPhase(pp,lockp) ;

	 return r ;
      } ;

      //----------------------------------------
      // Chagamma0:: confirm or push current PlayPhase
      // 	return True if new PlayPhase is pushed
   public:
      Bool confirmPushPlayPhase(PlayPhase::Type type,
				Int arg1 = 0,
				Int arg2 = 0,
				Int arg3 = 0,
				Bool lockp = True) {
	 Bool r ;
	 if(lockp) ppstack.lock() ;
	 
	 if(ppstack.isEmpty()) {
	    pushPlayPhase(type,arg1,arg2,arg3,False) ;
	    r = True ;
	 } else {
	    PlayPhase & pp = ppstack.top() ;
	    if(pp.type == type && pp.iarg1() == arg1 && pp.iarg2() == arg2 
	       && pp.iarg3() == arg3) {
	       r = False ;
	    } else {
	       pushPlayPhase(type,arg1,arg2,arg3,False) ;
	       r = True ;
	    }
	 }
	 if(lockp) ppstack.unlock() ;
	 return r ;
      } ;

      Bool confirmPushPlayPhase(PlayPhase::Type type,
				Flt arg1,
				Flt arg2 = 0.0,
				Flt arg3 = 0.0,
				Bool lockp = True) {
	 Bool r ;
	 if(lockp) ppstack.lock() ;
	 if(ppstack.isEmpty()) {
	    pushPlayPhase(type,arg1,arg2,arg3,False) ;
	    r = True ;
	 } else {
	    PlayPhase & pp = ppstack.top() ;
	    if(pp.type == type && pp.farg1() == arg1 && pp.farg2() == arg2 
	       && pp.farg3() == arg3) {
	       r = False ;
	    } else {
	       pushPlayPhase(type,arg1,arg2,arg3,False) ;
	       r = True ;
	    }
	 }
	 if(lockp) ppstack.unlock() ;
	 return r ;
      } ;
	    
      //----------------------------------------
      // Chagamma0:: confirm and pop current PlayPhase
      // 	return True if PlayPhase is poped
   public:
      Bool confirmPopPlayPhase(PlayPhase::Type type, Bool lockp = True) {
	 Bool r ;
	 if(lockp) ppstack.lock() ;
	 PlayPhase & pp = ppstack.top() ;
	 if(pp.type == type) {
	    popPlayPhase(False) ;
	    r = True ;
	 } else {
	    r = False ;
	 } 
	 if(lockp) ppstack.unlock() ;
	 return r ;
      } ;

      //----------------------------------------
      // Chagamma0:: virtual analyzeSensorSee()
   public:
      virtual Bool analyzeSensorSee() ;

      //----------------------------------------
      // Chagamma0:: virtual analyzeSensorHear() ;
   public:
      virtual Bool analyzeSensorHear() ;

      //----------------------------------------
      // Chagamma0:: for debug
   public:
      virtual Bool send() {
//	 ITK_MSG("sb:" << *this << btime().rtime) ;
//	 ITK_MSG("send:" << *this << TimeVal::now()) ;
	 return SimplePlayer::send() ;
      } ;


  //======================================================================
  // run Offline Coach

  int runOfflineCoach(int argc, char** argv) {
    OfflineCoach coach ;
    Bool r = coach.init(SHost,OffCoachPort,7.0,True) ;
    ITK_DBG(r) ;

    coach.forkSensor() ;

    coach.sendEye(True) ;
    coach.sendEar(True) ;
    coach.sendChangeMode(PLAYMODE(BeforeKickOff)) ;
   
    coach.cvSensor.lock() ;
    while(True) {
      coach.cvSensor.wait() ;
      ITK_MSG(TimeVal::now()) ;
      ITK_DBG(coach.recvsexp) ;
    }
  
    return 0 ;
  } ;

} ;


#endif // __Chagama_ChagamaBase_hxx__
