// -*- Mode:C++ -*-
//Header:
//File: NnPredictor.h
//Title: NnPredictor.h
//Author: NODA, Itsuki
//Date: 2001/08/01
//

//ModifyHistory:
// 2001/08/01: Start to create this file
//EndModifyHistory:

#if USE_NNN
#ifndef _saloo_NnPredictor_h_
#define _saloo_NnPredictor_h_
////////////////////////////////////////////////////////////////////////

#include <stl.h>
#include "SimplePlayer.h"
//#include "OnlineCoach.h"
//#include "OfflineCoach.h"
#include "WorldModel.h"

#include "MentalWorld.h"
#include "TrainSetBase.h"
#include "itk/Geo2D.h"

using namespace Itk ;

using namespace Nnn ;

namespace Saloo {

   //typedef Geo2D::Vec Vec ;

   //======================================================================
   // class Ball Tracker Train Set

   class BallTracker : public MwTrainSetBase {

	 //------------------------------------------------------------
	 // BallTracker:: for MwTrainSetBase
      public:
	 virtual const char * name() const { return "BallTracker" ; } ;
	 virtual Int sensorN() const { return 6 ; } ;
	 virtual Int actionN() const { return 2 ; } ;

	 //------------------------------------------------------------
	 // BallTracker:: scale parameter for values
      public:
	 static const Flt pscale = 0.02 ; /* scale for network pos input */
	 static const Flt vscale = 2.00 ; /* scale for network vel input */
	 static const Flt ascale = 0.01 ; /* scale for network acc input */
	 static const Flt dscale = 1.00 ; /* scale for network dir input */

	 //------------------------------------------------------------
	 // BallTracker:: member

	 //--------------------------------------------------
	 // BallTracker:: WorldModel
      public:
	 WorldModel * worldmodel_ ;
	 ITK_DEF_ACCESS(WorldModel & worldmodel(), { return *worldmodel_ ;}) ;

	 WorldModel & setWorldModel(WorldModel & wm) { 
	    worldmodel_ = &wm ;
	    return worldmodel() ;
	 } ;

	 //------------------------------------------------------------
	 // BallTracker:: cycle (dummy)
      public:
	 virtual void cycle(ValVector & sensor, ValVector & senMask,
			    ValVector & action, ValVector & actMask) {} ;

	 //------------------------------------------------------------
	 // BallTracker:: getFrame (override of definition of MwTrainSetBase)
      public:
	 virtual Frame & getFrame(Int time) {
	    Frame & f = frame(time) ;
	    WorldModel::Object & ball = worldmodel().estBall(time) ;

	    f.sensor().set(0.0,0,4) ;
	    f.senMask().set(0.0,0,4) ;

	    f.sensor()[4] = ball.pos().x ;
	    f.sensor()[5] = ball.pos().y ;

	    Flt conf = ((ball.posConf < 0.0000001) ? 0.0 
			: Min(ball.posConf * 100000,1.0)) ;

	    f.senMask()[4] = conf ;
	    f.senMask()[5] = conf ;

	    f.action().set(0.0) ;
	    f.actMask().set(0.0) ;

	    return f ;
	 } ;
   } ;

   //======================================================================
   // class BallPredictor

   class BallPredictor : public MentalWorld {

	 //------------------------------------------------------------
	 // BallPredictor:: constants
      public:
	 static char * const NetFileName = "saloo.net" ;
	 static const UInt ExpandN = 20 ;

	 //------------------------------------------------------------
	 // BallPredictor:: member
      public:
	 BallTracker ballTracker ;
	 
	 //------------------------------------------------------------
	 // BallPredictor:: constructor
      public:
	 //BallPredictor() { init() ; } ;

	 //------------------------------------------------------------
	 // BallPredictor:: setup
      public:
	 void setup(WorldModel & wm, char * const netfile = NetFileName) {
	    load(netfile) ;
	    expand(ExpandN) ;
	    ballTracker.setWorldModel(wm) ;
	 } ;
	    
	 //------------------------------------------------------------
	 // BallPredictor:: estimatePos
	 //		suppose balltracker know the pos before time-1
#ifdef NNN
      public:
	 Pos estimatePos(Int time, Int baseLen, UInt adaptN, Val beta) {
	    for(Int i = 0 ; i < baseLen ; i++) {
	       Int t = time-baseLen + i ;
	       WorldModel::Object & ball = 
		  ballTracker.worldmodel().estBall(t) ;

	       frame[i].sensor->inp()[4] = ball.pos().x ;
	       frame[i].sensor->inp()[5] = ball.pos().y ;
	       frame[i].senmask->set(0.0,0,4) ;
	       frame[i].senmask->set(0.99,4,6) ; // !!!!!
	       frame[i].action->inp().set(0.0) ;
	       frame[i].actmask->set(0.0) ;
	    }
	    frame[baseLen].sensor->inp().set(0.0) ;
	    frame[baseLen].senmask->set(0.0) ;
	    frame[baseLen].action->inp().set(0.0) ;
	    frame[baseLen].actmask->set(0.0) ;

	    propAdapt(baseLen+1, adaptN, beta) ;

	    //return Pos(frame[baseLen].expect->out()[0],
	    //	       frame[baseLen].expect->out()[1]) ;
	    return Pos(frame[baseLen].expect->out()[4],
	    	       frame[baseLen].expect->out()[5]) ;
	 } ;
#endif
   } ;
} ;
      








////////////////////////////////////////////////////////////////////////
#endif
#endif

